/***************************************************************************
                          board_2d.cpp  -  description
                             -------------------
    begin                : Fri Feb 28 2003
    copyright            : (C) 2003 by The Knights Project
    email                : knights-general@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "board_2d.moc"
#include "resource.h"
#include "logic.h"
#include "knightspixcache.h"
#include <kiconeffect.h>
#include <qpainter.h>
#include <qtimer.h>
#include <qstyle.h>
#include <qtooltip.h>

static int MAX_STEPS = 45;

///////////////////////////////////////
//
//  board_2d::constructor
//
///////////////////////////////////////
board_2d::board_2d(QWidget *parent, const char *name, resource *Rsrc, logic *Lgc  ) : board_base(parent,name,Rsrc,Lgc)
{
    init = true;
    premoveFrom = Null;
    premoveTo = Null;
    DragSprite = NULL;
    lastMoveWasDrag = false;
    cache = myResource->pixCache;
    sprites.setAutoDelete( true );

    /* Setup Pixmaps */
    drawBuffer.setOptimization( QPixmap::BestOptim );
    drawBuffer.resize( 16, 16);
    drawBuffer.fill();

    /* Setup children */
    messageBar          = new QLabel( this, "messageBar");
    whiteTimeLabel      = new QLabel( this, "whiteTimeLabel");
    blackTimeLabel      = new QLabel( this, "blackTimeLabel");
    notationBar         = new KComboBox( this, "notationBar");

    /* Init Message */
    messageBar->setAlignment( Qt::AlignAuto | Qt::AlignVCenter | Qt::SingleLine );
    messageBar->setFrameStyle( QFrame::ToolBarPanel | QFrame::Sunken );

    QColorGroup GroupWhite;
    QColorGroup GroupBlack;
    /* Init White Time Label */
    GroupWhite.setColor( QColorGroup::Background, myResource->COLOR_White );
    GroupWhite.setColor( QColorGroup::Foreground, myResource->COLOR_Black );
    whiteTimeLabel->setPalette( QPalette( GroupWhite, GroupWhite, GroupWhite ) );
    whiteTimeLabel->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
    whiteTimeLabel->setAlignment( Qt::AlignCenter | Qt::SingleLine );

    /* Init Black Time Label */
    GroupBlack.setColor( QColorGroup::Background, myResource->COLOR_Black );
    GroupBlack.setColor( QColorGroup::Foreground, myResource->COLOR_White );
    blackTimeLabel->setPalette( QPalette( GroupBlack, GroupBlack, GroupBlack ) );
    blackTimeLabel->setFrameStyle( QFrame::StyledPanel | QFrame::Sunken );
    blackTimeLabel->setAlignment( Qt::AlignCenter | Qt::SingleLine );

    /* Set Tooltips */
    QToolTip::add( whiteTimeLabel, i18n("White's Clock") );
    QToolTip::add( blackTimeLabel, i18n("Black's Clock") );

    connect( notationBar, SIGNAL( activated(int) ), this, SIGNAL( requestReview(int) ) );
    connect( this, SIGNAL( statusbarText( const QString& ) ), messageBar, SLOT( setText( const QString& ) ) );
    connect( cache, SIGNAL( themeChanged() ), this, SLOT( themeChange() ) );

    /* Setup self */
    setBackgroundMode( Qt::NoBackground );
    setStatusMessage( READY );
    setMinimumSize( 128, 128 );
    init = false;
}
///////////////////////////////////////
//
//  board_2d::destructor
//
///////////////////////////////////////
board_2d::~board_2d()
{
    delete messageBar;
    delete whiteTimeLabel;
    delete blackTimeLabel;
    delete notationBar;
}
///////////////////////////////////////
//
//  board_2d::coords
//
///////////////////////////////////////
QPoint board_2d::coords( const int &rank, const int &file )
{
    QPoint tmp;

    if( rank < 8 )
    {
        /* Standard Position */
        if( orientation == 0 )
        {
            tmp.setX( squareSize * file );
            tmp.setY( squareSize * ( 7 - rank ) );
        }
        else
        {
            tmp.setX( squareSize * ( 7 - file ) );
            tmp.setY( squareSize * rank );
        }
        if( myResource->ThemeBorder )
        {
            tmp.setX( tmp.x() + ( squareSize >> 1 ) );
            tmp.setY( tmp.y() + ( squareSize >> 1 ) );
        }
        tmp += rectBoard.topLeft();
    }
    else
    {
        /* Sideboard */
        int pos = ( rank * 8 ) + file;
        if( pos < 69 )
        {
            if( orientation == 0 )
            {
                tmp.setY( ( pos - 64 ) * sideSquareSize );
                tmp += rectWSide.topLeft();
            }
            else
            {
                tmp.setY( ( 4 - ( pos - 64 ) ) * sideSquareSize );
                tmp += rectBSide.topLeft();
            }
        }
        else
        {
            if( orientation == 0 )
            {
                tmp.setY( ( pos - 69 ) * sideSquareSize );
                tmp += rectBSide.topLeft();
            }
            else
            {
                tmp.setY( ( 4 - ( pos - 69 ) ) * sideSquareSize );
                tmp += rectWSide.topLeft();
            }
        }
    }
    return tmp;
}
///////////////////////////////////////
//
//      board_2d::position
//
///////////////////////////////////////
int board_2d::position( const QPoint &_point )
{
    QPoint point( _point );

    if( rectBoard.contains( point ) )
    {
        int file, rank;
        point -= rectBoard.topLeft();
        if( myResource->ThemeBorder )
        {
            point.setX( point.x() - ( squareSize >> 1 ) );
            point.setY( point.y() - ( squareSize >> 1 ) );
            if( ( point.x() < 0 ) || ( point.y() < 0 ) ||
                    ( point.x() > squareSize * 8 ) || ( point.y() > squareSize * 8 ) )
            {
                return Null;
            }
        }
        if( !orientation )
        {
            file = point.x() / squareSize;
            rank = 7 - ( point.y() / squareSize );
        }
        else
        {
            file = 7 - ( point.x() / squareSize );
            rank = point.y() / squareSize;
        }
        return ( ( rank << 3 ) + file );
    }
    if( usingSideboard )
    {
        if( rectWSide.contains( point ) )
        {
            point -= rectWSide.topLeft();
            if( orientation == 0 )
            {
                return ( 64 + ( point.y() / sideSquareSize ) );
            }
            else
            {
                return ( 69 + ( point.y() / sideSquareSize ) );
            }
        }
        else if( rectBSide.contains( point ) )
        {
            point -= rectBSide.topLeft();
            if( orientation == 0 )
            {
                return ( 69 + ( point.y() / sideSquareSize ) );
            }
            else
            {
                return ( 64 + ( point.y() / sideSquareSize ) );
            }
        }
    }
    return Null;
}
///////////////////////////////////////
//
//      board_2d::drawMove
//
///////////////////////////////////////
void board_2d::drawMove( const ChessMove &chessMove, const bool &reverse )
{
    signed char fromPtr, toPtr, takenPtr(Null);
    if( reverse )
    {
        fromPtr = ( chessMove.toRank << 3 ) + chessMove.toFile;
        toPtr = ( chessMove.fromRank << 3 ) + chessMove.fromFile;
    }
    else
    {
        fromPtr = ( chessMove.fromRank << 3 ) + chessMove.fromFile;
        toPtr = ( chessMove.toRank << 3 ) + chessMove.toFile;
    }

		/* Position where Man was taken != Target Position; ie. en passant */
		if( chessMove.ManTaken != Null )
		{
			takenPtr = myLogic->pointer( myLogic->chessman[ chessMove.ManTaken ].File, myLogic->chessman[ chessMove.ManTaken ].Rank );
		}
		else if( QString( chessMove.SAN ).contains( "x" ) )
		{
			takenPtr = toPtr;
		}

    /* Show Highlights */
    if( myResource->OPTION_Show_Last_Move )
    {
        myLogic->current[ fromPtr ].Note = NOTE_MOVE;
        if( takenPtr != Null )
        {
            myLogic->current[ toPtr ].Note = NOTE_ATTACK;
        }
        else
        {
            myLogic->current[ toPtr ].Note = NOTE_MOVE;
        }
    }

    /* Show Animation */
    if( myResource->OPTION_Animate_Moves && !lastMoveWasDrag && isVisible() )
    {
        QTimer::singleShot( 1, this, SLOT( updateVisuals() ) );
        Sprite *spritePtr = new Sprite;
        spritePtr->Steps = 1;
        spritePtr->Restore = false;
        spritePtr->POSITION_Origin = fromPtr;
        spritePtr->POSITION_Destination = toPtr;
        spritePtr->POSITION_TargetTaken = takenPtr;
        if( !reverse )
        {
            spritePtr->POINT_Origin = coords( chessMove.fromRank, chessMove.fromFile );
            spritePtr->POINT_Destination = coords( chessMove.toRank, chessMove.toFile );
            spritePtr->PIXMAP_Sprite = getChessman( spritePtr->POSITION_Destination );
        }
        else
        {
            spritePtr->POINT_Origin = coords( chessMove.toRank, chessMove.toFile );
            spritePtr->POINT_Destination = coords( chessMove.fromRank, chessMove.fromFile );
            spritePtr->PIXMAP_Sprite = getChessman( spritePtr->POSITION_Origin );
        }
        spritePtr->POINT_Current = spritePtr->POINT_Origin;
        spritePtr->PIXMAP_FlipFrame.resize( spritePtr->PIXMAP_Sprite.size() );
        sprites.append( spritePtr );
    }
    else
    {
        /* Draw this position only if we're not animating */
        drawVisuals( toPtr, HighlightFadeUp );
        if( takenPtr != Null )
        {
            drawVisuals( takenPtr, HighlightFadeUp );
        }
    }

    /* Draw the originating position */
    drawVisuals( fromPtr, HighlightFadeUp );

    if( QString( chessMove.SAN ).contains( "o-o", false ) )
    {
        /* This is a castle */
        ChessMove newMove;
        strcat( newMove.SAN, QString( "no" ).latin1() );
        newMove.fromRank = chessMove.fromRank;
        newMove.toRank = chessMove.toRank;
        newMove.ManTaken = Null;
        if( QString( chessMove.SAN ).contains( "o-o-o", false ) )
        {
            /* Queenside */
            newMove.fromFile = 0;
            newMove.toFile = 3;
        }
        else
        {
            /* Kingside */
            newMove.fromFile = 7;
            newMove.toFile = 5;
        }
        drawMove( newMove, reverse );
    }
    lastMoveWasDrag = false;
}
///////////////////////////////////////
//
//  board_2d::themeChange
//
///////////////////////////////////////
void board_2d::themeChange( void )
{
    QResizeEvent *event=0;
    resizeEvent( event );
}
///////////////////////////////////////
//
//  board_2d::resizeEvent
//
///////////////////////////////////////
void board_2d::resizeEvent( QResizeEvent* )
{
    /* Init */
    int tmp(0);
    int statusX(0);
    int size = 8 + ( 1 * ( myResource->ThemeBorder == true ) );
    int margin = QApplication::style().pixelMetric( QStyle::PM_DefaultFrameWidth );

    /* Get the proper height for the status bar */
    int statusHeight = notationBar->sizeHint().height();
    if( statusHeight < messageBar->sizeHint().height() )
    {
        statusHeight = messageBar->sizeHint().height();
    }
    if( statusHeight < whiteTimeLabel->sizeHint().height() )
    {
        statusHeight = whiteTimeLabel->sizeHint().height();
    }
    statusHeight += margin;

    /* Determine our need to display Sideboards */
    if( ( myLogic->variation() == GameType::BugHouse ) ||
            ( myLogic->variation() == GameType::CrazyHouse ) )
    {
        usingSideboard = true;
    }
    else
    {
        usingSideboard = false;
    }

    /* Get new size for each square on the board */
    int newWidth = width() / ( size + ( 2 * usingSideboard ) );
    int newHeight = ( height() - statusHeight - margin ) / size;
    if( newWidth < newHeight )
    {
        squareSize = newWidth;
    }
    else
    {
        squareSize = newHeight;
    }

    if( ( squareSize < IMAGE_MIN ) || ( squareSize > IMAGE_MAX ) || ( init == true ) )
    {
        return;
    }
    boardSize = squareSize * size;
    sideSquareSize = boardSize / 10;

    /* Resize the surface */
    drawBuffer.resize( width(), height() );

    /* Set Rectangles for all drawing areas */
    if( usingSideboard )
    {
        rectBoard.setRect( ( width() - boardSize - sideSquareSize - ( squareSize / 2 ) ) / 2, ( height() - boardSize - statusHeight - margin ) / 2, boardSize, boardSize );
        rectBSide.setRect( rectBoard.right() + ( squareSize / 2 ), rectBoard.top(), sideSquareSize, sideSquareSize * 5 );
        rectWSide.setRect( rectBoard.right() + ( squareSize / 2 ), rectBoard.top() + ( boardSize / 2 ), sideSquareSize, sideSquareSize * 5 );
    }
    else
    {
        rectBoard.setRect( ( width() - boardSize ) / 2, ( height() - boardSize - statusHeight - margin ) / 2, boardSize, boardSize );
    }

    /* Now paint the background */
    QPainter painter( &drawBuffer );
    if( cache->Background.isNull() )
    {
        painter.setBackgroundColor( paletteBackgroundColor() );
        painter.eraseRect( drawBuffer.rect() );
    }
    else
    {
        painter.drawTiledPixmap( drawBuffer.rect(), cache->Background );
    }

    /* Finished Painting */
    painter.end();

    /* Calculate granularity of animation */
    int inverseSize = IMAGE_MAX - squareSize;
    MAX_STEPS = inverseSize + 8;

    /* More init based on the new square size */
    int gridX( width() >> 3 );
    int statusY = drawBuffer.height() - margin - statusHeight;

    /* Do the Message box */
    tmp = gridX * 3;
    messageBar->setFixedSize( tmp - ( margin << 1 ), statusHeight );
    messageBar->move( margin, statusY );
    statusX = tmp;

    /* Do the White Time */
    tmp = gridX + ( gridX >> 1 );
    whiteTimeLabel->setFixedSize( tmp - margin, statusHeight );
    whiteTimeLabel->move( statusX, statusY );
    statusX += tmp;

    /* Do the Black Time */
    tmp = gridX + ( gridX >> 1 );
    blackTimeLabel->setFixedSize( tmp - margin, statusHeight );
    blackTimeLabel->move( statusX, statusY );
    statusX += tmp;

    /* Do the Notation Bar */
    tmp = gridX << 1;
    notationBar->setFixedSize( tmp - margin, statusHeight );
    notationBar->move( statusX, statusY );
    statusX += tmp;

    redrawAll();
}
///////////////////////////////////////
//
//      board_2d::redrawAll
//
///////////////////////////////////////
void board_2d::redrawAll( void )
{
    if( ( squareSize < IMAGE_MIN ) || ( squareSize > IMAGE_MAX ) )
        return;

    setCursor( myResource->CURSOR_Thinking );

    /* Set Orientation */
    orientation = myResource->OPTION_Board_Orientation;
    if( localArmy != WHITE )
        orientation = !orientation;
    if( flip )
        orientation = !orientation;

    /* Set Border */
    if( myResource->ThemeBorder )
    {
        QPixmap buffer;
        if( orientation )
        {
            cache->find( QString("orig_%1_flipborder").arg(squareSize), buffer );
        }
        else
        {
            cache->find( QString("orig_%1_border").arg(squareSize), buffer );
        }
        bitBlt( &drawBuffer, rectBoard.x(), rectBoard.y(),
                &buffer, 0, 0, buffer.width(), buffer.height(), Qt::CopyROP );
    }

    /* Redraw All Positions */
    register char tmp(0);
    if( usingSideboard )
    {
        while( tmp < 74 )
        {
            drawPosition( tmp++ );
        }
    }
    else
    {
        while( tmp < 64 )
        {
            drawPosition( tmp++ );
        }
    }

    if( myResource->ThemeBorder )
    {
        redrawLights();
    }

    repaint( false );
    setCursor( myResource->CURSOR_Standard );
}
///////////////////////////////////////
//
//      board_2d::redrawLights
//
///////////////////////////////////////
void board_2d::redrawLights( void )
{
    if( myResource->ThemeBorder )
    {
        QPoint black, white;
        int half = squareSize >> 1;
        int four = squareSize << 2;
        QPixmap temp;

        /* Get Positions */
        if( !orientation )
        {
            black.setX( 0 );
            black.setY( four );
            white.setX( 0 );
            white.setY( four + half );
        }
        else
        {
            black.setX( boardSize - half );
            black.setY( four + half );
            white.setX( boardSize - half );
            white.setY( four );
        }

        /* Clear the Old Lights... Needed for Transparent Textures */
        QPixmap border;
        if( orientation )
        {
            cache->find( QString("orig_%1_flipborder").arg(squareSize), border );
        }
        else
        {
            cache->find( QString("orig_%1_border").arg(squareSize), border );
        }
        QPainter painter( &drawBuffer );
        painter.drawPixmap( black.x() + rectBoard.left(), black.y() + rectBoard.top(), border,
                            black.x(), black.y(), half, half );
        painter.drawPixmap( white.x() + rectBoard.left(), white.y() + rectBoard.top(), border,
                            white.x(), white.y(), half, half );
        painter.end();

        black += rectBoard.topLeft();
        white += rectBoard.topLeft();

        if( myLogic->OnMove != WHITE )
        {
            cache->find( QString("orig_%1_on").arg(squareSize), temp );
            myBlit( black, &temp, temp.rect() );
            cache->find( QString("orig_%1_off").arg(squareSize), temp );
            myBlit( white, &temp, temp.rect() );
        }
        else
        {
            cache->find( QString("orig_%1_off").arg(squareSize), temp );
            myBlit( black, &temp, temp.rect() );
            cache->find( QString("orig_%1_on").arg(squareSize), temp );
            myBlit( white, &temp, temp.rect() );
        }
    }
}
///////////////////////////////////////
//
//      board_2d::drawPosition
//
///////////////////////////////////////
void board_2d::drawPosition( const int &pos )
{
    drawPosition( pos, 1.0f );
}
void board_2d::drawPosition( const int &pos, const float &highlightAlpha )
{
    int rank = pos >> 3;
    int file = pos % 8;
    QPixmap buffer;
    QString cacheName;
    int mySquareSize;

    if( ( pos < 0 ) || ( pos > ( 63 + ( 10 * ( usingSideboard ) ) ) ) )
    {
        return;
    }

    // This is just for convienance
    Position thePosition = myLogic->current[pos];

    /* Get the square size we'll be working with */
    if( pos < 64 )
    {
        mySquareSize = squareSize;
    }
    else
    {
        mySquareSize = sideSquareSize;
    }

    /*
        Build the cache ref name
    */
    cacheName = QString::number( mySquareSize );
    if( pos < 64 )
    {
        /* Use standard colors for board */
        if( color( rank, file ) )
        {
            cacheName += "L";
        }
        else
        {
            cacheName += "D";
        }
    }
    else
    {
        /* Use simple color lookup for Sideboard */
        if( pos % 2 )
        {
            cacheName += "L";
        }
        else
        {
            cacheName += "D";
        }
    }
    if( !paused )
    {
        switch( thePosition.Note )
        {
        case NOTE_SELECT:
        case NOTE_HIGHLIGHT:
            cacheName += QString("S_%1_").arg( highlightAlpha );
            break;
        case NOTE_MOVE:
        case NOTE_CASTLE:
        case NOTE_PAWN_DOUBLE:
            cacheName += QString("M_%1_").arg( highlightAlpha );
            break;
        case NOTE_ATTACK:
        case NOTE_ENPASSANT:
            cacheName += QString("A_%1_").arg( highlightAlpha );
            break;
        default:
            cacheName += " ";
        }
        if( ( thePosition.ManPtr != Null ) && ( !isSprite( pos ) ) )
        {
            if( ( pos < 64 ) || ( myLogic->chessman[ thePosition.ManPtr ].File > 0 ) )
            {
                if( myLogic->chessman[thePosition.ManPtr].Army == WHITE )
                {
                    cacheName += "W";
                }
                else
                {
                    cacheName += "B";
                }
                switch( myLogic->chessman[thePosition.ManPtr].Type )
                {
                case King:
                    cacheName += "K";
                    break;
                case Queen:
                    cacheName += "Q";
                    break;
                case Bishop:
                    cacheName += "B";
                    break;
                case Knight:
                    cacheName += "N";
                    break;
                case Rook:
                    cacheName += "R";
                    break;
                case Pawn:
                    cacheName += "P";
                    break;
                default:
                    break;
                }
            }
            if( ( pos == premoveFrom ) || ( pos == premoveTo ) )
            {
                cacheName += "t"; // This means it's transparent
            }
        }
    }
    else
    {
        cacheName += " ";
    }

    /* Look in the cache for our image */
    if( !cache->find( cacheName, buffer ) )
    {
        /*
          Cache miss
          Draw the pixmap
        */
        if( pos < 64 )
        {
            if( color( rank, file ) )
            {
                cache->find( QString("orig_%1_light").arg(mySquareSize), buffer );
            }
            else
            {
                cache->find( QString("orig_%1_dark").arg(mySquareSize), buffer );
            }
        }
        else
        {
            /* Use simple color lookup for Sideboard */
            if( pos % 2 )
            {
                cache->find( QString("orig_%1_light").arg(mySquareSize), buffer );
            }
            else
            {
                cache->find( QString("orig_%1_dark").arg(mySquareSize), buffer );
            }
        }

        if( !paused )
        {
            QString highlight;
            switch( thePosition.Note )
            {
            case NOTE_HIGHLIGHT:        // Fall Through
            case NOTE_SELECT:
                highlight += QString("orig_%1_select_%2").arg(mySquareSize).arg(highlightAlpha);
                break;
            case NOTE_MOVE:                             // Fall Through
            case NOTE_CASTLE:                   // Fall Through
            case NOTE_PAWN_DOUBLE:
                highlight += QString("orig_%1_move_%2").arg(mySquareSize).arg(highlightAlpha);
                break;
            case NOTE_ATTACK:                   // Fall Through
            case NOTE_ENPASSANT:
                highlight += QString("orig_%1_attack_%2").arg(mySquareSize).arg(highlightAlpha);
                break;
            }

            if( !highlight.isEmpty() )
            {
                QPixmap temp;
                cache->find( highlight, temp );
                bitBlt( &buffer, 0, 0, &temp, 0, 0, -1, -1, Qt::CopyROP, false);
            }

            if( !isSprite(pos) )
            {
                if( ( pos < 64 ) || ( myLogic->chessman[ thePosition.ManPtr ].File > 0 ) )
                {
                    QPixmap chessman = getChessman( pos );
                    bitBlt( &buffer, 0, 0, &chessman, 0, 0, -1, -1, Qt::CopyROP, false);
                }
            }

        }

        /* Now add this pixmap to the cache */
        cache->add
        ( cacheName, buffer );
    }


    if( myResource->OPTION_Show_Coord )
    {
        /* Draw Coordinates */
        drawCoords( &buffer, pos );
    }

    if( pos > 63 )
    {
        /* Draw Piece Count */
        drawCount( &buffer, myLogic->chessman[ thePosition.ManPtr ].File );
    }

    /* Draw to the main board */
    myBlit( coords( rank, file ), &buffer, buffer.rect() );
}
///////////////////////////////////////
//
//      board_2d::drawCoords
//
///////////////////////////////////////
void board_2d::drawCoords( QPixmap *pic, const int &pos )
{
    QPainter painter;
    QString letter;
    int themeSize = squareSize - 4;
    int targetRank, targetFile;

    if( !orientation )
    {
        targetRank = 0;
        targetFile = 0;
    }
    else
    {
        targetRank = 7;
        targetFile = 7;
    }
    /* Draw Rank */
    if( ( pos >> 3 ) == targetRank )
    {
        letter = QString( QString("abcdefgh").at( pos % 8 ) );
        painter.begin( pic );
        painter.setFont( myResource->FONT_Standard );
        painter.setPen( myResource->COLOR_Notation_Shadow );
        painter.drawText( 3, 3, themeSize, themeSize,
                          Qt::AlignRight | Qt::AlignBottom, letter );
        painter.setPen( myResource->COLOR_Notation );
        painter.drawText( 2, 2, themeSize, themeSize,
                          Qt::AlignRight | Qt::AlignBottom, letter );
        painter.end();
    }
    /* Draw File */
    if( ( pos % 8 ) == targetFile )
    {
        letter = QString( QString("12345678").at( pos >> 3 ) );
        painter.begin( pic );
        painter.setFont( myResource->FONT_Standard );
        painter.setPen( myResource->COLOR_Notation_Shadow );
        painter.drawText( 3, 3, themeSize, themeSize,
                          Qt::AlignLeft | Qt::AlignTop, letter );
        painter.setPen( myResource->COLOR_Notation );
        painter.drawText( 2, 2, themeSize, themeSize,
                          Qt::AlignLeft | Qt::AlignTop, letter );
        painter.end();
    }
}
///////////////////////////////////////
//
//      board_2d::drawCount
//
///////////////////////////////////////
void board_2d::drawCount( QPixmap *pic, const int &num )
{
    QPainter painter;
    QString letter;
    int themeSize = sideSquareSize - 4;

    letter = QString::number( num );
    painter.begin( pic );
    painter.setFont( myResource->FONT_Standard );
    painter.setPen( myResource->COLOR_Notation_Shadow );
    painter.drawText( 3, 3, themeSize, themeSize,
                      Qt::AlignRight | Qt::AlignBottom, letter );
    painter.setPen( myResource->COLOR_Notation );
    painter.drawText( 2, 2, themeSize, themeSize,
                      Qt::AlignRight | Qt::AlignBottom, letter );
    painter.end();
}
///////////////////////////////////////
//
//      board_2d::mouseReleaseEvent
//
///////////////////////////////////////
void board_2d::mouseReleaseEvent( QMouseEvent *event )
{
    event->accept();

    if( DragSprite != NULL )
    {
        /* Destroy any sprites being dragged */
        int tmp = DragSprite->POSITION_Origin;
        myBlit( DragSprite->POINT_LastUpdate,
                &DragSprite->PIXMAP_FlipFrame,
                DragSprite->PIXMAP_FlipFrame.rect() );
        sprites.removeRef( DragSprite );
        DragSprite = NULL;
        drawPosition( tmp );
        lastMoveWasDrag = true;
    }

    if(event->button() == Qt::LeftButton)
    {
        emit leftClick( position( event->pos() ) );
    }
    if(event->button() == Qt::RightButton)
    {
        emit rightClick( position( event->pos() ) );
    }
}
///////////////////////////////////////
//
//      board_2d::mousePressEvent
//
///////////////////////////////////////
void board_2d::mousePressEvent( QMouseEvent *event )
{
    pressPoint = event->pos();
    event->accept();
}
///////////////////////////////////////
//
//      board_2d::mouseMoveEvent
//
///////////////////////////////////////
void board_2d::mouseMoveEvent( QMouseEvent *event )
{
    event->accept();
    QPoint currentPos = event->pos();
    if( DragSprite == NULL )
    {
        if( event->state() & Qt::LeftButton )
        {
            if( abs( pressPoint.x() - currentPos.x() ) + abs( pressPoint.y() - currentPos.y() ) > 6 )
            {
                /* Begin Dragging a piece */
                DragSprite = new Sprite;
                sprites.append( DragSprite );
                DragSprite->POINT_Origin = pressPoint;
                DragSprite->POSITION_Origin = position( pressPoint );
                emit leftClick( DragSprite->POSITION_Origin ); // Tell match that we just clicked on this piece

                if( myLogic->current[ DragSprite->POSITION_Origin ].Note != NOTE_SELECT )
                {
                    /* The selection didn't take.. back out. */
                    sprites.removeRef( DragSprite );
                    DragSprite = NULL;
                    drawPosition( position( pressPoint ) );
                    return;
                }

                /* Get the piece image and store it in dragPix */
                DragSprite->PIXMAP_Sprite = getChessman( DragSprite->POSITION_Origin );
                DragSprite->PIXMAP_FlipFrame.resize( DragSprite->PIXMAP_Sprite.size() );
                DragSprite->Restore = false;
            }
            else
            {
                /* Not enough dragging */
                return;
            }
        } // End ( event->state() & Qt::LeftButton )
        else
        {
            return; /* No dragging. Most events should end up here */
        }
    }

    int halfSize = squareSize >> 1;
    DragSprite->POINT_Current.setX( currentPos.x() - halfSize );
    DragSprite->POINT_Current.setY( currentPos.y() - halfSize );
    drawSprites();
}
///////////////////////////////////////
//
//      board_2d::wheelEvent
//
///////////////////////////////////////
void board_2d::wheelEvent( QWheelEvent *event )
{
    event->accept();
    if( event->delta() > 0 )
    {
        prevNotation();
    }
    if( event->delta() < 0 )
    {
        nextNotation();
    }
}
///////////////////////////////////////
//
//      board_2d::prevNotation
//
///////////////////////////////////////
void board_2d::prevNotation( void )
{
    int tmp( notationBar->currentItem() - 1 );
    emit requestReview( tmp ); // Do this before the bounds checking in case this is examine mode
    if( tmp < 0 )
        return;
    notationBar->setCurrentItem( tmp );
}
///////////////////////////////////////
//
//      board_2d::nextNotation
//
///////////////////////////////////////
void board_2d::nextNotation( void )
{
    int tmp( notationBar->currentItem() + 1 );
    emit requestReview( tmp ); // Do this before the bounds checking in case this is examine mode
    if( tmp > ( notationBar->count() - 1 ) )
        return;
    notationBar->setCurrentItem( tmp );
}
///////////////////////////////////////
//
//      board_2d::getChessman
//
///////////////////////////////////////
QPixmap board_2d::getChessman( const int &pos )
{
    QPixmap temp;
    QString fetch;
    register char type, army;

    if( pos == premoveTo )
    {
        type = myLogic->chessman[myLogic->current[premoveFrom].ManPtr].Type;
        army = myLogic->chessman[myLogic->current[premoveFrom].ManPtr].Army;
    }
    else
    {
        type = myLogic->chessman[myLogic->current[pos].ManPtr].Type;
        army = myLogic->chessman[myLogic->current[pos].ManPtr].Army;
    }

    if( pos < 64 )
    {
        fetch = QString("orig_%1_").arg(squareSize);
    }
    else
    {
        fetch = QString("orig_%1_").arg(sideSquareSize);
    }
    if( army == WHITE )
    {
        fetch += "w";
    }
    else
    {
        fetch += "b";
    }
    switch( type )
    {
    case King:
        fetch += "k";
        break;
    case Queen:
        fetch += "q";
        break;
    case Bishop:
        fetch += "b";
        break;
    case Knight:
        fetch += "n";
        break;
    case Rook:
        fetch += "r";
        break;
    case Pawn:
        fetch += "p";
        break;
    default:
        break;
    }

    /* Set Alpha */
    if( ( pos == premoveFrom ) || ( pos == premoveTo ) )
    {
        fetch += "_0.5";
    }
    cache->find( fetch, temp );
    return temp;
}
///////////////////////////////////////
//
//      board_2d::setPremovePositions
//
///////////////////////////////////////
void board_2d::setPremovePositions( const int &posF, const int &posT )
{
    premoveFrom = posF;
    premoveTo = posT;
    if( ( posF != Null ) && ( posT != Null ) )
    {
        myLogic->current[posF].Note = NOTE_NONE;
        myLogic->current[posT].Note = NOTE_NONE;
        drawPosition( posF );
        drawPosition( posT );
    }
}
///////////////////////////////////////
//
//      board_2d::paintEvent
//
///////////////////////////////////////
void board_2d::paintEvent( QPaintEvent *event )
{
    QRegion reg = event->region();

    QPainter painter( this );
    painter.setClipRegion( reg );
    painter.drawPixmap( 0, 0, drawBuffer );

    painter.end();
}
///////////////////////////////////////
//
//      board_2d::drawSprites
//
///////////////////////////////////////
void board_2d::drawSprites( void )
{
    int index;
    Sprite *spritePtr;

    /* Remove all sprites from the pixmap in reverse order */
    QPainter painter( &drawBuffer );

    for( index = (signed int)sprites.count() - 1; index > -1; index-- )
    {
        spritePtr = sprites.at(index);
        if( spritePtr->Restore == true )
        {
            painter.drawPixmap( spritePtr->POINT_LastUpdate, spritePtr->PIXMAP_FlipFrame );
            update( QRect( spritePtr->POINT_LastUpdate, spritePtr->PIXMAP_FlipFrame.size() ) );
        }
    }
    /* Redraw all sprites */
    for( index = 0; index < (signed int)sprites.count(); index++ )
    {
        spritePtr = sprites.at(index);

        if( ( spritePtr == DragSprite ) || ( spritePtr->Steps < MAX_STEPS ) )
        {
            /* Obtain next flip frame */
            bitBlt( &spritePtr->PIXMAP_FlipFrame,
                    0,
                    0,
                    &drawBuffer,
                    spritePtr->POINT_Current.x(),
                    spritePtr->POINT_Current.y(),
                    spritePtr->PIXMAP_FlipFrame.width(),
                    spritePtr->PIXMAP_FlipFrame.height(),
                    Qt::CopyROP );
            if( spritePtr->Restore == false )
            {
                spritePtr->Restore = true;
            }
            /* Redraw Sprite */
            painter.drawPixmap( spritePtr->POINT_Current, spritePtr->PIXMAP_Sprite );
            update( QRect( spritePtr->POINT_Current, spritePtr->PIXMAP_Sprite.size() ) );
            spritePtr->POINT_LastUpdate = spritePtr->POINT_Current;
        }
        else
        {
            /* Animation finished */
            int origin = spritePtr->POSITION_Origin;
            int destination = spritePtr->POSITION_Destination;
            int target = spritePtr->POSITION_TargetTaken;
            sprites.removeRef( spritePtr );
            drawPosition( origin );
            drawVisuals( destination, HighlightFadeUp );
            if( destination != target )
            {
                drawVisuals( target, HighlightFadeUp );
            }
            index = -1;
        }
    }
    painter.end();
}
///////////////////////////////////////
//
//      board_2d::updateVisuals
//
///////////////////////////////////////
void board_2d::updateVisuals( void )
{
    bool callAgain = false;

    /* Update effects */
    if( myResource->OPTION_Fade && effects.count() )
    {
        for( int index = 0; index < (signed int)effects.count(); index++ )
        {
            callAgain = true;
            bool remove
                = false;
            Effect *effectPtr = effects.at( index );

            if( effectPtr->Fade_Start < effectPtr->Fade_Stop )
            {
                effectPtr->Fade_Current += effectPtr->Fade_Step;
                if( effectPtr->Fade_Current > effectPtr->Fade_Stop )
                {
                    effectPtr->Fade_Current = effectPtr->Fade_Stop;
                    remove
                        = true;
                }
            }
            else
            {
                effectPtr->Fade_Current -= effectPtr->Fade_Step;
                if( effectPtr->Fade_Current < effectPtr->Fade_Stop )
                {
                    effectPtr->Fade_Current = effectPtr->Fade_Stop;
                    remove
                        = true;
                }
            }

            switch( effectPtr->Type )
            {
            case HighlightFadeUp:
            case HighlightFadeDown:
                drawPosition( effectPtr->Position, effectPtr->Fade_Current );
                break;
            case ChessmanFadeUp:
            case ChessmanFadeDown:
                break;
            default:
                break;
            }

            if( remove
                  )
                {
                    effects.removeRef( effectPtr );
                    index = -1;
                }
        }
    }

    /* Take care of sprites */
    if( myResource->OPTION_Animate_Moves && sprites.count() )
    {
        for( int index = 0; index < (signed int)sprites.count(); index++ )
        {
            Sprite *spritePtr = sprites.at( index );
            if( spritePtr != DragSprite )
            {
                callAgain = true;
                if( spritePtr->Steps < MAX_STEPS )
                {
                    double factor = ( 1.0 / (double)MAX_STEPS ) * (double)spritePtr->Steps;
                    int newX = (int)( ( spritePtr->POINT_Destination.x() - spritePtr->POINT_Origin.x() ) * factor );
                    int newY = (int)( ( spritePtr->POINT_Destination.y() - spritePtr->POINT_Origin.y() ) * factor );
                    spritePtr->POINT_Current = spritePtr->POINT_Origin + QPoint( newX, newY );
                    spritePtr->Steps += sprites.count();
                }
            }
        }
        drawSprites();
    }

    if( callAgain )
    {
        QTimer::singleShot( 1, this, SLOT( updateVisuals() ) );
    }
}
///////////////////////////////////////
//
//      board_2d::isSprite
//
///////////////////////////////////////
bool board_2d::isSprite( const int &pos )
{
    for( unsigned int index = 0; index < sprites.count(); index++ )
    {
        Sprite *tmpSprite = sprites.at( index );
        if( tmpSprite->POSITION_Origin == pos )
        {
            return true;
        }
    }
    return false;
}
///////////////////////////////////////
//
//  board_2d::drawVisuals
//
///////////////////////////////////////
void board_2d::drawVisuals( const int &pos, const EffectType &et )
{
    if( myResource->OPTION_Fade && ( myLogic->current[ pos ].Note != NOTE_NONE ) )
    {
        QTimer::singleShot( 1, this, SLOT( updateVisuals() ) );
        Effect *effectPtr = new Effect;
        effectPtr->Position = pos;
        effectPtr->Type = et;
        switch( et )
        {
        case HighlightFadeUp:
        case ChessmanFadeUp:
            effectPtr->Fade_Start = 0.00f;
            effectPtr->Fade_Current = 0.00f;
            effectPtr->Fade_Step = 0.015f;
            effectPtr->Fade_Stop = 1.0f;
            break;
        case HighlightFadeDown:
        case ChessmanFadeDown:
            effectPtr->Fade_Start = 1.00f;
            effectPtr->Fade_Current = 1.00f;
            effectPtr->Fade_Step = 0.015f;
            effectPtr->Fade_Stop = 0.0f;
            break;
        }
        effects.append( effectPtr );
    }
    else
    {
        drawPosition( pos );
    }
}
///////////////////////////////////////
//
//      board_2d::myBlit
//
///////////////////////////////////////
void board_2d::myBlit( const QPoint &dp, const QPaintDevice *src, const QRect &sr )
{
    bitBlt( &drawBuffer, dp.x(), dp.y(), src, sr.x(), sr.y(), sr.width(), sr.height(), Qt::CopyROP );
    update( QRect( dp.x(), dp.y(), sr.width(), sr.height() ) );
}
///////////////////////////////////////
//
//      board_2d::setClocks
//
///////////////////////////////////////
void board_2d::setClocks( const QString &white, const QString &black )
{
    whiteTimeLabel->setText( white );
    blackTimeLabel->setText( black );
}
///////////////////////////////////////
//
//      board_2d::setNotation
//
///////////////////////////////////////
void board_2d::setNotation( QStringList *list, int index )
{
    QStringList::Iterator IT;
    int count(0);

    notationBar->clear();
    for( IT = list->begin(); IT != list->end(); ++IT )
    {
        notationBar->insertItem( (*IT), count++ );
    }
    if( index == -1 )
    {
      notationBar->setCurrentItem( --count );
    }
    else
    {
      notationBar->setCurrentItem( index );
    }
    delete list;
}
///////////////////////////////////////
//
//      board_2d::sizeHint
//
///////////////////////////////////////
QSize board_2d::sizeHint( void ) const
{
    /* Default size should be one quarter of the screen real estate */
    QDesktopWidget *desktop = QApplication::desktop();
    return QSize( desktop->height() / 2, ( desktop->height() / 2 ) + notationBar->sizeHint().height() );
}
