/***************************************************************************
                          board_base.cpp  -  description
                             -------------------
    begin                : Fri Feb 28 2003
    copyright            : (C) 2003 by The Knights Team
    email                : knights-general@lists.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "board_base.moc"
#include "resource.h"
#include "logic.h"
#include "pgn.h"
#include <kmessagebox.h>
#include <kpopupmenu.h>

board_base::board_base(QWidget *parent, const char *name, resource *Rsrc, logic *Lgc ) : QWidget(parent,name)
{
    /* Setup Variables */
    myResource = Rsrc;
    myLogic = Lgc;

    paused = false;
    flip = false;
    localArmy = WHITE;
}
board_base::~board_base()
{}
///////////////////////////////////////
//
//      board_base::setPaused
//
///////////////////////////////////////
void board_base::setPaused( const bool &state )
{
    paused = state;
    redrawAll();
}
///////////////////////////////////////
//
//      board_base::setLocalArmy
//
///////////////////////////////////////
void board_base::setLocalArmy( const bool &state )
{
    if( state == localArmy )
        return;
    localArmy = state;
    redrawAll();
}
///////////////////////////////////////
//
//      board_base::setStatusMessage
//
///////////////////////////////////////
void board_base::setStatusMessage( const int &ID, const QString &MSG )
{
    /*
                Game comments as specified in http://www.schachprobleme.de/chessml/faq/pgn/
                Section 10: Numeric Annotation Glyphs
    */
    if( ( ID & COMMENT ) == COMMENT )
    {
        emit statusbarText( pgn::getNAG( ID - COMMENT ) );
        return;
    }
    /* Knights specific messages */
    switch( ID )
    {
        /* Error Codes */
    case BOOK_ERROR_1:
        emit statusbarText( i18n( "Error with white engine" ) );
        KMessageBox::sorry( this,
                            i18n("You selected %1 to play white,\nbut it can only be used as a book engine.\nPlease select another engine to play white.").arg( MSG ),
                            i18n("White Engine Problem") );
        break;
    case BOOK_ERROR_2:
        emit statusbarText( i18n( "Error with white book engine" ) );
        KMessageBox::sorry( this,
                            i18n("You selected %1 to play white's book,\nbut it can only be used as a regular engine.\nPlease select another engine to play white's book.").arg( MSG ),
                            i18n("White Book Engine Problem") );
        break;
    case BOOK_ERROR_3:
        emit statusbarText( i18n( "Error with black engine" ) );
        KMessageBox::sorry( this,
                            i18n("You selected %1 to play black,\nbut it can only be used as a book engine.\nPlease select another engine to play black.").arg( MSG ),
                            i18n("Black Engine Problem") );
        break;
    case BOOK_ERROR_4:
        emit statusbarText( i18n( "Error with black book engine" ) );
        KMessageBox::sorry( this,
                            i18n("You selected %1 to play black's book,\nbut it can only be used as a regular engine.\nPlease select another engine to play black's book.").arg( MSG ),
                            i18n("Black Book Engine Problem") );
        break;
    case ENGINE_DIED_ERROR:
        emit statusbarText( i18n( "The computer opponent assigned to play %1 has crashed" ).arg( MSG ) );
        break;
    case LOAD_ERROR:
        emit statusbarText( i18n( "There was an error while loading the file" ) );
        break;
    case SAVE_ERROR:
        emit statusbarText( i18n( "There was an error while saving the file" ) );
        break;
        /* Standard Codes */
    case LOAD_OK:
        emit statusbarText( i18n( "Loading complete" ) );
        break;
    case SAVE_OK:
        emit statusbarText( i18n( "Saving complete" ) );
        break;
    case READING_FILE:
        emit statusbarText( i18n( "Reading File" ) );
        break;
    case NO_MOVE_WHILE_REVIEW:
        emit statusbarText( i18n( "Can not move a chessman while reviewing the match" ) );
        break;
    case ILLEGAL_MOVE:
        emit statusbarText( i18n( "Illegal Move" ) );
        break;
    case WHITE_TURN:
        emit statusbarText( i18n( "White's turn" ) );
        break;
    case BLACK_TURN:
        emit statusbarText( i18n( "Black's turn" ) );
        break;
    case WHITE_WIN:
        emit statusbarText( i18n( "White wins" ) );
        break;
    case BLACK_WIN:
        emit statusbarText( i18n( "Black wins" ) );
        break;
    case WHITE_CHECKMATE:
        emit statusbarText( i18n( "Checkmate, White wins" ) );
        break;
    case BLACK_CHECKMATE:
        emit statusbarText( i18n( "Checkmate, Black wins" ) );
        break;
    case WHITE_RESIGN:
        emit statusbarText( i18n( "White resigns" ) );
        break;
    case BLACK_RESIGN:
        emit statusbarText( i18n( "Black resigns" ) );
        break;
    case WHITE_FLAG:
        emit statusbarText( i18n( "White's time elapsed" ) );
        break;
    case BLACK_FLAG:
        emit statusbarText( i18n( "Black's time elapsed" ) );
        break;
    case WHITE_CALL_FLAG:
        emit statusbarText( i18n( "Black was called on time, White wins" ) );
        break;
    case BLACK_CALL_FLAG:
        emit statusbarText( i18n( "White was called on time, Black wins" ) );
        break;
    case GAME_DRAW:
        emit statusbarText( i18n( "Draw match" ) );
        break;
    case GAME_50_MOVES:
        emit statusbarText( i18n( "50 moves rule, draw match" ) );
        break;
    case WAITING:
        emit statusbarText( i18n( "Starting computer players, please wait" ) );
        break;
    case PAUSED:
        emit statusbarText( i18n( "Match paused" ) );
        break;
    case WHITE_DRAW_OFFER:
        emit statusbarText( i18n( "White has offered a draw" ) );
        break;
    case BLACK_DRAW_OFFER:
        emit statusbarText( i18n( "Black has offered a draw" ) );
        break;
    case LOST_CONTACT:
        emit statusbarText( i18n( "Lost contact with opponent" ) );
    case READY:
    default:
        emit statusbarText( i18n( "Ready" ) );
        break;
    }
}
///////////////////////////////////////
//
//      board_base::recvCMD
//
///////////////////////////////////////
void board_base::recvCMD( const Command &cmdIn )
{
    Command cmd = cmdIn;

    switch( cmd.getCommand() )
    {
        /* CMD_Move */
    case MENU_ORIENTATION:
        flip = !flip;
        redrawAll();
        break;

    default:
        emit sendCMD( cmdIn );
        break;
    }
}
///////////////////////////////////////
//
//      board_base::requestWidgetFeatureList
//
///////////////////////////////////////
void board_base::requestWidgetFeatureList( void )
{
  emit internalReqWFL();

  /* Send Menu */
  QString caption = i18n( "&Match" );
  emit fetchMenu( myMenu, caption );
}
///////////////////////////////////////
//
//      board_base::setMenu
//
///////////////////////////////////////
void board_base::setMenu( KPopupMenu *menu )
{
    myMenu = menu;
}
///////////////////////////////////////
//
//      board_base::drawPosition
//
///////////////////////////////////////
void board_base::drawPosition( const int& )
{}
///////////////////////////////////////
//
//      board_base::setPremovePositions
//
///////////////////////////////////////
void board_base::setPremovePositions( const int&, const int& )
{}
///////////////////////////////////////
//
//      board_base::redrawLights
//
///////////////////////////////////////
void board_base::redrawLights( void )
{}
///////////////////////////////////////
//
//      board_base::setNotation
//
///////////////////////////////////////
void board_base::setNotation( QStringList* list, int index )
{}
