/***************************************************************************
                          match.cpp  -  description
                             -------------------
    begin                : Mon Jul 2 2001
    copyright            : (C) 2003 by Troy Corbin Jr.
    email                : tcorbin@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <qregexp.h>
#include <kmessagebox.h>
#include <kfiledialog.h>
#include <kpopupmenu.h>
#include "match.moc"
#include "board_base.h"
#include "board_2d.h"
#include "audio.h"
#include "tabmanager.h"
#include "core.h"
#include "matchhistory.h"

#define WHITE_INPUT Record->Param->type(WHITE)
#define BLACK_INPUT Record->Param->type(BLACK)

///////////////////////////////////////
//
//      match::match
//
///////////////////////////////////////
match::match( core *parent, match_param *param, resource *Rsrc ) : QObject(parent,"match")
{
    myResource = Rsrc;

    /* Init Children */
    Record = new pgn( myResource, param );
    Logic = new logic( myResource, param );
    Clock = new chessclock( this, myResource );

    createMenus();

    Board = new board_2d( 0, "Board", myResource, Logic );
    connect( Board, SIGNAL( leftClick(int) ), this, SLOT( slot_Select(int) ) );
    connect( Board, SIGNAL( rightClick(int) ), this, SLOT( slot_Preview(int) ) );
    connect( this, SIGNAL( setStatusBar( const int&, const QString& ) ), Board, SLOT( setStatusMessage( const int&, const QString& ) ) );
    connect( Board, SIGNAL( requestReview(int) ), this, SLOT( review(int) ) );
    connect( Record, SIGNAL( requestReview(int) ), this, SLOT( review(int) ) );
    connect( Board, SIGNAL( requestClose( bool& ) ), this, SLOT( destroyMatch( bool& ) ) );
    connect( Board, SIGNAL( sendCMD( const Command& ) ), this, SLOT( recvCMD( const Command& ) ) );
    connect( this, SIGNAL( updateWidgetFeatureList( const WFL& ) ), Board, SIGNAL( updateWidgetFeatureList( const WFL& ) ) );
    connect( Board, SIGNAL( internalReqWFL() ), this, SLOT( requestWidgetFeatureList( void ) ) );
    Board->setMenu( m_match );

    Record->init();
    //Logic->init( GameType::CrazyHouse ); // Bughouse for testing
    Logic->init( param->getVariation() );
    Clock->Reset();
    clearSelections();

    /* Init Variables */
    Draw_Offered[WHITE] = false;
    Draw_Offered[BLACK] = false;
    Loading = false;
    Paused = false;
    JustMoved = false;
    preMoved = false;
    Current = false;
    loadTimer = 0;
    StatusCode = READY;
    StatusMessage = QString::null;
    ICSGameMode = Null;
    myFeatureList.ContextMenu = true;
    myFeatureList.ReqPermissionForClose = true;
    myFeatureList.EnableSave = false;
    myFeatureList.EnableSaveAs = true;
    myFeatureList.EnablePrint = true;

    /* Connect Signals and Slots */
    connect( param, SIGNAL( valuesChanged() ), this, SLOT( parseMatchParam() ) );
    connect( Record, SIGNAL( processMove(ChessMove) ), this, SLOT( move(ChessMove) ) );
    connect( Record, SIGNAL( processSpecial() ), this, SLOT( loadSpecial() ) );
    connect( Clock, SIGNAL( flagFell(const bool) ), this, SLOT( slot_flagFell(const bool) ) );

    /* Insert this match into the Tabs system */
    myResource->tabManager->addTab( Board, Record->caption() );
    parseMatchParam();
}
///////////////////////////////////////
//
//      match::~match
//
///////////////////////////////////////
match::~match()
{
    delete Clock;
    delete Record;
    delete Logic;
    delete m_draw;
    delete m_match;
}
///////////////////////////////////////
//
//      match::destroyMatch
//
///////////////////////////////////////
void match::destroyMatch( bool &status )
{
    /* Save if needed */
    if( myFeatureList.EnableSave && !myResource->ShutdownMode )
    {
        if( save( true, false ) == false )
        {
            status = false;
            return;
        }
    }
    status = true;

    /* Tell core to disconnect us from any IOs */
    ((core*)parent())->clearMatch( myID );
}
///////////////////////////////////////
//
//      match::save
//
///////////////////////////////////////
bool match::save( const bool Prompt, const bool SaveAs )
{
    int answer;
    QString URL;

    if( Prompt )
    {
        /* Prompt for the save */
        myResource->ConfigWrite();
        answer = KMessageBox::warningYesNoCancel( 0,
                 i18n("Would you like to save this match?"),
                 i18n("Save Match?"),
                 QString::null,
                 QString::null,
                 QString("PromptForSaving") );
        myResource->ConfigRead();
        if( answer == KMessageBox::Cancel )
        {
            return false;
        }
        if( answer == KMessageBox::No )
        {
            return true;
        }
    }
    /* Get the last URL we used */
    URL = url();
    if( ( URL.isEmpty() ) || ( SaveAs ) )
    {
        /* Get a new URL */
        URL = KFileDialog::getSaveFileName( QString::null, myResource->FileFilter, 0, i18n( "Save Match..." ) );
        if( URL.isNull() )
        {
            return false;
        }
        if( URL.findRev( ".pgn", -1, false ) == -1 )
        {
            URL += ".pgn";
        }
    }

    Record->whiteTime = Clock->getCentiseconds(WHITE);
    Record->blackTime = Clock->getCentiseconds(BLACK);
    if( Record->save( URL ) == false )
    {
        setStatusMessage( SAVE_ERROR );
        return false;
    }
    setStatusMessage( SAVE_OK );
    setModified( false );
    return true;
}
///////////////////////////////////////
//
//      match::parseMatchParam
//
///////////////////////////////////////
void match::parseMatchParam( void )
{
    /* Set menus to default values */
    m_match->setItemEnabled( MENU_DRAW, false );
    m_match->setItemEnabled( MENU_RESIGN, false );
    m_match->setItemEnabled( MENU_MOVE_NOW, false );
    m_match->setItemEnabled( MENU_HINT, false );
    m_match->setItemEnabled( MENU_PONDER, false );
    m_match->setItemEnabled( MENU_PAUSE, false );

    switch( BLACK_INPUT )
    {
    case PLAYERLOCAL:
        Board->setLocalArmy( BLACK );
        m_match->setItemEnabled( MENU_DRAW, true );
        m_match->setItemEnabled( MENU_RESIGN, true );
        break;
    case PLAYERPC:
        m_match->setItemEnabled( MENU_MOVE_NOW, true );
        m_match->setItemEnabled( MENU_HINT, true );
        m_match->setItemEnabled( MENU_PONDER, true );
        m_match->setItemEnabled( MENU_PAUSE, true );
        break;
    }
    switch( WHITE_INPUT )
    {
    case PLAYERLOCAL:
        Board->setLocalArmy( WHITE );
        m_match->setItemEnabled( MENU_DRAW, true );
        m_match->setItemEnabled( MENU_RESIGN, true );
        break;
    case PLAYERPC:
        m_match->setItemEnabled( MENU_MOVE_NOW, true );
        m_match->setItemEnabled( MENU_HINT, true );
        m_match->setItemEnabled( MENU_PONDER, true );
        m_match->setItemEnabled( MENU_PAUSE, true );
        break;
    }
    myResource->tabManager->changeCaption( Board, Record->caption() );
}
///////////////////////////////////////
//
//      match::tick
//
///////////////////////////////////////
void match::tick(void)
{
    Clock->Tick();
    Board->setClocks( Clock->whiteClock, Clock->blackClock );

    /* Update the Match menu */
    if( Clock->Flag[WHITE] )
    {
        if( BLACK_INPUT == PLAYERLOCAL )
            m_match->setItemEnabled( MENU_CALL_FLAG, true );
        else
            m_match->setItemEnabled( MENU_CALL_FLAG, false );
    }
    if( Clock->Flag[BLACK] )
    {
        if( WHITE_INPUT == PLAYERLOCAL )
            m_match->setItemEnabled( MENU_CALL_FLAG, true );
        else
            m_match->setItemEnabled( MENU_CALL_FLAG, false );
    }
}
///////////////////////////////////////
//
//      match::createMenus
//
///////////////////////////////////////
void match::createMenus( void )
{
    m_draw = new KPopupMenu( 0, "m_draw");
    // MENU_OFFER_DRAW
    m_draw->insertItem( i18n( "&Offer Draw" ), this, SLOT( recvMenuAction(int) ), 0, MENU_OFFER_DRAW );
    //  m_draw->setWhatsThis( MENU_OFFER_DRAW, i18n( "Clicking this will inform your opponent that you are willing draw the match." ) );
    // MENU_ACCEPT_DRAW
    m_draw->insertItem( i18n( "&Accept Draw" ), this, SLOT( recvMenuAction(int) ), 0, MENU_ACCEPT_DRAW );
    //  m_draw->setWhatsThis( MENU_ACCEPT_DRAW, i18n( "Clicking this will accept a draw offered by your opponent." ) );
    // MENU_REJECT_DRAW
    m_draw->insertItem( i18n( "&Reject Draw" ), this, SLOT( recvMenuAction(int) ), 0, MENU_REJECT_DRAW );
    //  m_draw->setWhatsThis( MENU_REJECT_DRAW, i18n( "Clicking this will reject a draw offered by your opponent." ) );
    // MENU_IGNORE_DRAW
    m_draw->insertItem( i18n( "&Ignore Draw" ), this, SLOT( recvMenuAction(int) ), 0, MENU_IGNORE_DRAW );
    //  m_draw->setWhatsThis( MENU_IGNORE_DRAW, i18n( "Clicking this will ignore future draw offers from your opponent." ) );

    m_match = new KPopupMenu( 0, "m_match");
    m_match->setCheckable(true);
    // MENU_DRAW
    m_match->insertItem( i18n( "&Draw" ), m_draw, MENU_DRAW );
    m_match->setItemEnabled( MENU_DRAW, false );
    // MENU_RETRACT
    m_match->insertItem( QIconSet( myResource->LoadIcon( QString("undo"), KIcon::Small ) ),
                         i18n( "&Retract Move" ), this, SLOT( recvMenuAction(int) ), 0, MENU_RETRACT );
    m_match->setItemEnabled( MENU_RETRACT, false );
    //  m_match->setWhatsThis( MENU_RETRACT, i18n( "Select this to retract your last move." ) );
    // MENU_RESIGN
    m_match->insertItem( i18n( "Resign" ), this, SLOT( recvMenuAction(int) ), 0, MENU_RESIGN );
    m_match->setItemEnabled( MENU_RESIGN, false );
    //  m_match->setWhatsThis( MENU_RESIGN, i18n( "Use this to concede the match to your opponent." ) );
    // MENU_CALL_FLAG
    m_match->insertItem( i18n( "&Call Flag" ), this, SLOT( recvMenuAction(int) ), 0, MENU_CALL_FLAG );
    //  m_match->setWhatsThis( MENU_CALL_FLAG, i18n( "Use this to declare the match over, due to your opponent being out of time." ) );
    m_match->setItemEnabled( MENU_CALL_FLAG, false );
    // MENU_HINT
    m_match->insertItem( i18n( "&Hint" ), this, SLOT( recvMenuAction(int) ), 0, MENU_HINT );
    m_match->setItemEnabled( MENU_HINT, false );
    //  m_match->setWhatsThis( MENU_HINT, i18n( "This will ask your opponent for a hint." ) );
    // MENU_MOVE_NOW
    m_match->insertItem( i18n( "Move &Now" ), this, SLOT( recvMenuAction(int) ), 0, MENU_MOVE_NOW );
    m_match->setItemEnabled( MENU_MOVE_NOW, false );
    //  m_match->setWhatsThis( MENU_MOVE_NOW, i18n( "Clicking this option will force your opponent to move immediately." ) );
    // MENU_ORIENTATION
    m_match->insertItem( i18n( "&Flip View" ), this, SLOT( recvMenuAction(int) ), Key_F2, MENU_ORIENTATION );
    m_match->setItemEnabled( MENU_ORIENTATION, true );
    //  m_match->setWhatsThis( MENU_ORIENTATION, i18n( "This will rotate the chessboard's orientation by 180 degrees." ) );
    // MENU_PONDER
    m_match->insertItem( i18n( "&Ponder" ), this, SLOT( recvMenuAction(int) ), 0, MENU_PONDER );
    m_match->setItemChecked( MENU_PONDER, myResource->OPTION_Ponder );
    m_match->setItemEnabled( MENU_PONDER, false );
    //  m_match->setWhatsThis( MENU_PONDER, i18n( "This toggles your opponent's ability to think while it's your turn." ) );
    m_match->insertSeparator();
    // MENU_PAUSE
    m_match->insertItem( QIconSet( myResource->LoadIcon( QString("player_pause"), KIcon::Small ) ),
                         i18n( "Pause" ), this, SLOT( recvMenuAction(int) ), Key_F12, MENU_PAUSE );
    m_match->setItemEnabled( MENU_PAUSE, false );
    //  m_match->setWhatsThis( MENU_PAUSE, i18n( "Select this to pause the clock for this match." ) );
}
///////////////////////////////////////
//
//      match::clearSelections
//
///////////////////////////////////////
void match::clearSelections( void )
{
    int tmp;

    if( ICSGameMode == Null )
    {
        for( tmp = 0; tmp < _ARRAY_SIZE_; tmp++ )
        {
            if( Logic->current[tmp].Note != NOTE_NONE )
            {
                Logic->current[tmp].Note = NOTE_NONE;
                Board->drawPosition( tmp );
            }
        }
    }
    else
    {
        for( tmp = 0; tmp < _ARRAY_SIZE_; tmp++ )
        {
            Logic->current[tmp].Note = NOTE_NONE;
            Board->drawPosition( tmp );
        }
    }
}
///////////////////////////////////////
//
//      match::slot_flagFell
//
///////////////////////////////////////
void match::slot_flagFell( const bool Army )
{
    if( Army == WHITE )
    {
        emit setStatusBar( WHITE_FLAG );
        Record->history->add( Clock->getCentiseconds( WHITE ), Clock->getCentiseconds( BLACK ) );
        if( ( myResource->OPTION_Auto_Call_Flag ) && ( WHITE_INPUT != PLAYERLOCAL ) && ( BLACK_INPUT == PLAYERLOCAL ) )
        {
            emit sendCMD( Command( myID, CMD_Black_Called_Flag ) );
            if( WHITE_INPUT != PLAYERTCP )
            {
                recvCMD( Command( myID, CMD_Black_Called_Flag ) );
            }
        }
    }
    else
    {
        emit setStatusBar( BLACK_FLAG );
        Record->history->add( Clock->getCentiseconds( WHITE ), Clock->getCentiseconds( BLACK ) );
        if( ( myResource->OPTION_Auto_Call_Flag ) && ( BLACK_INPUT != PLAYERLOCAL ) && ( WHITE_INPUT == PLAYERLOCAL ) )
        {
            emit sendCMD( Command( myID, CMD_White_Called_Flag ) );
            if( BLACK_INPUT != PLAYERTCP )
            {
                recvCMD( Command( myID, CMD_White_Called_Flag ) );
            }
        }
    }
}
///////////////////////////////////////
//
//      match::setPaused
//
///////////////////////////////////////
void match::setPaused( const bool State )
{
    Paused = State;
    Board->setPaused( State );
    if( Paused )
    {
        myResource->tabManager->changeCaption( Board, i18n("%1 - Paused").arg( Record->caption() ) );
        Clock->Pause();
        emit setStatusBar( PAUSED );
    }
    else
    {
        myResource->tabManager->changeCaption( Board, Record->caption() );
        Clock->Resume();
        if( onMove() == WHITE )
            emit setStatusBar( WHITE_TURN );
        else
            emit setStatusBar( BLACK_TURN );
    }
}
///////////////////////////////////////
//
//      match::requestWidgetFeatureList
//
///////////////////////////////////////
void match::requestWidgetFeatureList()
{
  emit updateWidgetFeatureList( myFeatureList );
}
///////////////////////////////////////
//
//      match::inputOnMove
//
///////////////////////////////////////
char match::inputOnMove( bool reverse )
{
    if( reverse )
    {
        if( Logic->OnMove == BLACK )
            return WHITE_INPUT;
        if( Logic->OnMove == WHITE )
            return BLACK_INPUT;
    }
    else
    {
        if( Logic->OnMove == WHITE )
            return WHITE_INPUT;
        if( Logic->OnMove == BLACK )
            return BLACK_INPUT;
    }
    return Null;
}
///////////////////////////////////////
//
//      match::input
//
///////////////////////////////////////
char match::input( char army )
{
    return Record->Param->type(army);
}
///////////////////////////////////////
//
//      match::load
//
///////////////////////////////////////
bool match::load( const QString URL, const int pos )
{
    bool Result;
    if( !Record->open( URL ) )
    {
        return false;
    }
    Loading = true;
    Result = Record->load( pos );
    if( Record->tags.find("FEN") )
    {
        // We must set up the FEN position
        Logic->setBoardFromFen( Record->tags.find("FEN") );
        sendCMD( Command( myID, CMD_Set_Board, Record->tags.find("FEN") ) );
    }

    Clock->setInfiniteTime( true );
    while( Loading )
    {
      /*
          All this is for a match we just finished loading
      */
      if( !Record->loadNext() )
      {
        Loading = false;
        emit sendCMD( Command( myID, CMD_Play ) );
        if( Record->tags.find( "Result" ) == "*" )
        {
          /* Unfinished Game */
          Clock->Set( Record->whiteTime, Record->blackTime, onMove() );
          Record->refresh();
        }
        else
        {
          /* Finished Match */
          if( Record->tags.find( "Result" ) == "1/2-1/2" )
            emit setStatusBar( GAME_DRAW );
          else if( Record->tags.find( "Result" ) == "1-0" )
            emit setStatusBar( WHITE_WIN );
          else if( Record->tags.find( "Result" ) == "0-1" )
            emit setStatusBar( BLACK_WIN );
          Record->refresh();
          review( 0 );
        }
      }
    }
    parseMatchParam();
    return Result;
}
///////////////////////////////////////
//
//      match::loadSpecial
//
///////////////////////////////////////
void match::loadSpecial( void )
{
    /*
                As pointless as this function may seem ( in light of these variables and more
                being setup in tick(), it is required to initialize a Chess Engine and feed it
                the moves from a saved game so you can play it again later.
    */
    emit sendCMD( Command( myID, CMD_New_Players ) );
}
///////////////////////////////////////
//
//      match::url
//
///////////////////////////////////////
QString match::url( void )
{
    if( !Record->CurrentURL.isEmpty() )
        return Record->CurrentURL;
    if( myResource->OPTION_Reuse_PGN )
        return myResource->PGN_Filename;
    return QString::null;
}
///////////////////////////////////////
//
//      match::slot_Select
//
///////////////////////////////////////
void match::slot_Select( int position )
{
    bool preMoving(false);
    int tmp, selected(Null);
    register char army;

    if( Paused )
    {
        return; // Duh
    }

    if( ( position < 0 ) || ( position > _ARRAY_SIZE_ ) )
    {
        return;
    }

    /* Clear all non-SELECT notes */
    for( tmp = 0; tmp < _ARRAY_SIZE_; tmp++ )
    {
        if( Logic->current[tmp].Note != NOTE_NONE )
        {
            if( Logic->current[tmp].Note != NOTE_SELECT )
            {
                Logic->current[tmp].Note = NOTE_NONE;
                Board->drawPosition( tmp );
            }
            else
            {
                selected = tmp; // We need this later
            }
        }
    }

    /* Check to make sure it's our turn to select. */
    if( !( inputOnMove() & PLAYERLOCAL ) )
    {
        if( ( inputOnMove(true) == PLAYERLOCAL ) && ( myResource->OPTION_Premove ) )
        {
            preMoving = true;
        }
        else
        {
            return;
        }
    }

    /* If you left click on a selected square, it becomes unselected. */
    if( Logic->current[position].Note == NOTE_SELECT )
    {
        Logic->current[position].Note = NOTE_NONE;
        drawPosition( position );
        playSound( Sound::Select );
        return;
    }

    /* Clicking on sideboards with no pieces get rejected. */
    if( ( position > 63 ) && ( Logic->chessman[position].File < 1 ) )
    {
        return;
    }

    /* If you premoved and then click somewhere else, the premove is cleared. */
    if( preMoved )
    {
        preMoved = false;
        playSound( Sound::Select );
        Board->setPremovePositions( Null, Null );
        Board->drawPosition( Logic->pointer( preMove.fromFile, preMove.fromRank ) );
        Board->drawPosition( Logic->pointer( preMove.toFile, preMove.toRank ) );
    }

    /* Check to see if there is already a selected square. */
    if( selected != Null )
    {
        /*
                        If there is already a selected square, and you just clicked on another of your men,
                        we'll change your selection.
        */
        if( Logic->current[position].ManPtr != Null )
        {
            if( Logic->chessman[ Logic->current[selected].ManPtr ].Army == Logic->chessman[ Logic->current[position].ManPtr ].Army )
            {
                Logic->current[selected].Note = NOTE_NONE;
                Logic->current[position].Note = NOTE_SELECT;
                Board->drawPosition( selected );
                drawPosition( position );
                if( myResource->OPTION_Auto_Preview == true )
                    slot_Preview( position );
                else
                    playSound( Sound::Select );
                return;
            }
        }
        /*
                        If there is already a selected square, but this one you just clicked
                        isn't one of your chessmen, you must want to move a piece.
        */
        if( ( Record->currentIndex != ( Record->history->count( MatchHistory::Move ) - 1 ) ) && ( Record->currentIndex ) )
        {
            emit setStatusBar( NO_MOVE_WHILE_REVIEW );
            return;
        }
        if( preMoving )
        {
            Command::clearMove( &preMove );
            preMove.fromRank = Logic->current[selected].Rank;
            preMove.fromFile = Logic->current[selected].File;
            preMove.toRank = Logic->current[position].Rank;
            preMove.toFile = Logic->current[position].File;
            preMoved = true;
            playSound( Sound::Select );
            Board->setPremovePositions( selected, position );
        }
        else
        {
            Command::clearMove( &chessMove );
            chessMove.fromRank = Logic->current[selected].Rank;
            chessMove.fromFile = Logic->current[selected].File;
            chessMove.toRank = Logic->current[position].Rank;
            chessMove.toFile = Logic->current[position].File;
            JustMoved = true;
            if( move() == false )
            {
                Logic->current[selected].Note = NOTE_SELECT;
                drawPosition( selected );
            }
        }
        return;
    }
    tmp = Logic->current[position].ManPtr;

    /* You can't select an empty square. */
    if( tmp == Null )
    {
        return;
    }
    army = Logic->chessman[ tmp ].Army;

    /* You can't select your enemy */
    if( ( army != onMove() ) && ( preMoving == false ) )
    {
        return;
    }

    /* If your clicking on one of your chessmen, you can select it. */
    clearSelections();
    if( ( army == WHITE ) && ( WHITE_INPUT & PLAYERLOCAL ) )
    {
        Logic->current[position].Note = NOTE_SELECT;
    }
    if( ( army == BLACK ) && ( BLACK_INPUT & PLAYERLOCAL ) )
    {
        Logic->current[position].Note = NOTE_SELECT;
    }
    drawPosition( position );
    if( myResource->OPTION_Auto_Preview == true )
    {
        slot_Preview( position );
    }
    else
    {
        /* Play the select sound if you selected a square & you didn't preview */
        if( Logic->current[position].Note == NOTE_SELECT )
        {
            playSound( Sound::Select );
        }
    }
}
///////////////////////////////////////
//
//      match::slot_Preview
//
///////////////////////////////////////
void match::slot_Preview( int position )
{
    int tmp;

    if( Paused )
    {
        return;
    }
    if( ( position < 0 ) || ( position > _ARRAY_SIZE_ ) )
    {
        return;
    }

    clearSelections();
    tmp = Logic->current[position].ManPtr;
    if( tmp == Null )
        return;

    /* Clicking on sideboards with no pieces get rejected. */
    if( ( position > 63 ) && ( Logic->chessman[tmp].File < 1 ) )
    {
        return;
    }

    Logic->hashMoves( tmp );

    /* If your clicking on one of your chessmen, you can select it. */
    if( ( Logic->chessman[tmp].Army == Logic->OnMove ) && ( Record->Param->type( Logic->OnMove ) & PLAYERLOCAL ) )
    {
        Logic->current[position].Note = NOTE_SELECT;
        playSound( Sound::Select );
    }
    for( tmp = 0;tmp < _ARRAY_SIZE_;tmp++ )
    {
        if( Logic->current[tmp].Note != NOTE_NONE )
        {
            Board->drawPosition( tmp );
        }
    }
    drawPosition( 0 );
}
///////////////////////////////////////
//
//      match::recvCMD
//
///////////////////////////////////////
void match::recvCMD( const Command &cmd )
{
    if( cmd.getID() != myID )
    {
        /* Move isn't for Me */
        //return;
    }
    QString genericString; // Can't allocate in switches.
    MatchHistory::MatchEvent event; // Can't allocate in switches.
    ChessMove newMove = cmd.getMove();


    switch( cmd.getCommand() )
    {
        /* CMD_Move */
    case CMD_Move:
        if( newMove.ICS_Mode == Null )
        {
            /* Not an ICS Move, use standard move command */
            move( newMove );
        }
        else
        {
            /* Use ICS Move syntax */
            if( ( Logic->OnMove == -1 ) && ( newMove.ICS_Mode < ICS_Examine ) )
            {
                /* Do NOT accept a setBoard on a concluded match */
                break;
            }
            ICSGameMode = newMove.ICS_Mode;
            Logic->OnMove = newMove.ICS_OnMove;

            if( ICSGameMode == ICS_Movelist )
            {
                /* Recieving Movelist */
                Logic->chessMove = newMove;
                Logic->parseSAN();
                JustMoved = false;
            }
            else
            {
                /* Recieving Standard Move */
                Logic->setBoard( cmd.getData(), newMove.ICS_PawnPushFile );
                Logic->MoveCounter = newMove.ICS_MoveCounter;
                Logic->chessMove = newMove;
                Logic->parseCAN( newMove.ICS_OnMove );
            }
            chessMove = Logic->chessMove;
            /* Run through the Move function */
            if( !JustMoved )
            {
                if( Record->history->count( MatchHistory::Move ) && ( cmd.getData() == Record->history->last( MatchHistory::Move ).wholeBoard ) )
                {
                    /* User Refreshed */
                    retract();
                }
                move();
            }
            else
            {
                JustMoved = false;
                Logic->OnMove = !Logic->OnMove;
            }

            Clock->Set( cmd.getWhiteTime(), cmd.getBlackTime(), !newMove.ICS_OnMove );
            if( newMove.ICS_ClockTicking )
            {
                Clock->Resume();
            }
            else
            {
                Clock->Pause();
            }
        }
        break;

        /* CMD_Offer_Draw */
    case CMD_Offer_Draw:
        if( cmd.getData() == "W" )
        {
            Draw_Offered[WHITE] = 2;
            emit setStatusBar( WHITE_DRAW_OFFER );
        }
        else
        {
            Draw_Offered[BLACK] = 2;
            emit setStatusBar( BLACK_DRAW_OFFER );
        }
        playSound( Sound::DrawOffer );
        break;

        /* CMD_Illegal */
    case CMD_Illegal:
        retract();
        Board->setNotation( Record->notation() );
        break;

        /* Lost Contact */
    case CMD_Lost_Contact:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( LOST_CONTACT );
        Record->tags.replace( "Result", "*" );
        Record->tags.replace( "Termination", "Lost Contact with Opponent" );
        break;

        /* Result White */
    case CMD_Result_White:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( WHITE_CHECKMATE );
        Record->tags.replace( "Result", "1-0" );
        Record->tags.replace( "Termination", "Normal" );
        break;

        /* Result Black */
    case CMD_Result_Black:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( BLACK_CHECKMATE );
        Record->tags.replace( "Result", "0-1" );
        Record->tags.replace( "Termination", "Normal" );
        break;

        /* Result Draw */
    case CMD_Result_Draw:
        Clock->Pause();
        Logic->OnMove = -1;
        if( Logic->MoveCounter == 50 )
        {
            emit setStatusBar( GAME_50_MOVES );
        }
        else
        {
            emit setStatusBar( GAME_DRAW );
        }
        Record->tags.replace( "Result", "1/2-1/2" );
        break;

        /* White Resign */
    case CMD_White_Resign:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( WHITE_RESIGN );
        Record->tags.replace( "Result", "0-1" );
        Record->tags.replace( "Termination", "White Resigned" );
        break;

        /* Black Resign */
    case CMD_Black_Resign:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( BLACK_RESIGN );
        Record->tags.replace( "Result", "1-0" );
        Record->tags.replace( "Termination", "Black Resigned" );
        break;

        /* White Called Flag */
    case CMD_White_Called_Flag:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( WHITE_CALL_FLAG );
        Record->tags.replace( "Result", "1-0" );
        Record->tags.replace( "Termination", "Time Forfeit" );
        break;

        /* Black Called Flag */
    case CMD_Black_Called_Flag:
        Clock->Pause();
        Logic->OnMove = -1;
        emit setStatusBar( BLACK_CALL_FLAG );
        Record->tags.replace( "Result", "0-1" );
        Record->tags.replace( "Termination", "Time Forfeit" );
        break;

        /* Hint */
    case CMD_Hint:
        genericString = cmd.getData().section( '\n', 1 );
        strcpy( Logic->chessMove.SAN, genericString.latin1() );
        strcpy( Logic->chessMove.CAN, genericString.latin1() );
        clearSelections();
        Logic->hashMoves( Logic->current[ Logic->pointer( Logic->chessMove.fromFile, Logic->chessMove.fromRank ) ].ManPtr );
        if( Logic->parseCAN( onMove() ) != true )
        {
            Logic->parseSAN();
        }
        Board->drawMove( Logic->chessMove );
        KMessageBox::information( Board, cmd.getData(), i18n("Hint") );
        break;

        /* Tell User */
    case CMD_Tell_User:
        KMessageBox::information( Board, cmd.getData(), i18n( "Tell User" ) );
        break;

        /* Tell User Error */
    case CMD_Tell_User_Error:
        KMessageBox::sorry( Board, cmd.getData(), i18n( "Tell User Error" ) );
        break;

        /* Receive Piece */
    case CMD_Receive_Piece:
        drawPosition( Logic->recievePiece( newMove ) );
        Record->history->add( newMove, Clock->getCentiseconds( WHITE ), Clock->getCentiseconds( BLACK ) );
        break;

        /* Sideboard All */
    case CMD_Sideboard_All:
        Logic->recieveSideboard( cmd.getData() );
        Board->drawPosition( 64 );
        Board->drawPosition( 65 );
        Board->drawPosition( 66 );
        Board->drawPosition( 67 );
        Board->drawPosition( 68 );
        Board->drawPosition( 69 );
        Board->drawPosition( 70 );
        Board->drawPosition( 71 );
        Board->drawPosition( 72 );
        Board->drawPosition( 73 );
        // Update last move in history to include Sideboard
        event = Record->history->last();
        event.wholeBoard = Logic->board();
        Record->history->undo();
        Record->history->add( event );
        break;

    case MENU_PRINT:
        print();
        break;

    case MENU_SAVE:
        save( false, false );
        break;

    case MENU_SAVEAS:
        save( false, true );
        break;

    case MENU_PAUSEALL:
        recvMenuAction( MENU_PAUSE );
        break;

    default:
        break;
    }
}
///////////////////////////////////////
//
//      match::recvMenuAction
//
///////////////////////////////////////
void match::recvMenuAction( int cmd )
{
    switch( cmd )
    {
        /* Pause */
    case MENU_PAUSE:
        setPaused( !Paused );
        if( Paused )
        {
            sendCMD( Command( myID, CMD_Pause ) );
        }
        else
        {
            sendCMD( Command( myID, CMD_Resume ) );
        }
        break;

        /* Ponder */
    case MENU_PONDER:
        myResource->OPTION_Ponder = 1 - myResource->OPTION_Ponder;
        m_match->setItemChecked(MENU_PONDER, myResource->OPTION_Ponder );
        if( myResource->OPTION_Ponder )
        {
            sendCMD( Command( myID, CMD_Ponder ) );
        }
        else
        {
            sendCMD( Command( myID, CMD_No_Pondering ) );
        }
        break;

        /* Hint */
    case MENU_HINT:
        emit sendCMD( Command( myID, CMD_Hint ) );
        emit sendCMD( Command( myID, CMD_UCI_Hint, centiseconds( WHITE ), centiseconds( BLACK ), Record->notation(false)->join(QString(" ")) ) );
        break;

        /* Move Now */
    case MENU_MOVE_NOW:
        sendCMD( Command( myID, CMD_MoveNow ) );
        break;

        /* Retract */
    case MENU_RETRACT:
        sendCMD( Command( myID, CMD_Retract_Move ) );
        retract();
        retract();
        /* Set the statusbar */
        if( Logic->OnMove == WHITE )
        {
            setStatusMessage( WHITE_TURN );
        }
        else
        {
            setStatusMessage( BLACK_TURN );
        }
        /* Reset the menu option */
        if( ( inputOnMove() == PLAYERLOCAL ) &&
                ( inputOnMove(true) == PLAYERPC ) &&
                Record->history->count( MatchHistory::Move ) )
        {
            m_match->setItemEnabled( MENU_RETRACT, true );
        }
        else
        {
            m_match->setItemEnabled( MENU_RETRACT, false );
        }
        break;

        /* Call Flag */
    case MENU_CALL_FLAG:
        if( WHITE_INPUT == PLAYERLOCAL )
        {
            sendCMD( Command( myID, CMD_White_Called_Flag ) );
            recvCMD( Command( myID, CMD_White_Called_Flag ) );
        }
        if( BLACK_INPUT == PLAYERLOCAL )
        {
            sendCMD( Command( myID, CMD_Black_Called_Flag ) );
            recvCMD( Command( myID, CMD_Black_Called_Flag ) );
        }
        break;

        /* Resign */
    case MENU_RESIGN:
        if( ( WHITE_INPUT == PLAYERLOCAL ) && ( onMove() == WHITE ) )
        {
            sendCMD( Command( myID, CMD_White_Resign ) );
            recvCMD( Command( myID, CMD_White_Resign ) );
        }
        if( ( BLACK_INPUT == PLAYERLOCAL ) && ( onMove() == BLACK ) )
        {
            sendCMD( Command( myID, CMD_Black_Resign ) );
            recvCMD( Command( myID, CMD_Black_Resign ) );
        }
        break;

        /* Reject Draw */
    case MENU_REJECT_DRAW:
        sendCMD( Command( myID, CMD_Reject_Draw ) );
        break;

        /* Offer / Accept Draw */
    case MENU_OFFER_DRAW: // Fall Through
    case MENU_ACCEPT_DRAW:
        if( WHITE_INPUT == PLAYERLOCAL )
        {
            recvCMD( Command( myID, CMD_Offer_Draw, QString("W") ) );
        }
        if( BLACK_INPUT == PLAYERLOCAL )
        {
            recvCMD( Command( myID, CMD_Offer_Draw, QString("B") ) );
        }
        sendCMD( Command( myID, CMD_Offer_Draw ) );
        break;

        /* Flip View */
    case MENU_ORIENTATION:
        Board->recvCMD( Command( myID, MENU_ORIENTATION ) );
        break;

    default:
        break;
    }
}
///////////////////////////////////////
//
//      match::move
//
///////////////////////////////////////
bool match::move( ChessMove newMove )
{
    /* Clear selections from the board */
    clearSelections();

    Logic->chessMove = newMove;
    if( Logic->parseCAN( onMove() ) != true )
    {
        if( Logic->parseSAN() != true )
        {
            kdWarning() << "match::move #" << myID << ": Recieved bad data: CAN: " << newMove.CAN << " SAN: " << newMove.SAN << endl;
            return false;
        }
    }
    chessMove = Logic->chessMove;
    JustMoved = true;
    return move();
}
bool match::move( void )
{
    int soundType( Sound::Move );
    int result(Null);
    int manPtr(Null);
    QString SAN;


    /* Bring the display up to the latest move before we proceed */
    if( ICSGameMode == Null )
    {
        if( Record->currentIndex != ( Record->history->count( MatchHistory::Move ) - 1 ) )
        {
            review( Record->history->count( MatchHistory::Move ) - 1 );
        }
    }

    /* Clear selections from the board */
    clearSelections();

    /* Build Pointers */
    int fromPtr = ( chessMove.fromRank << 3 ) + chessMove.fromFile;

    /* Skip all this if we're being called due to normal server moves */
    if( ( JustMoved == true ) || ( ICSGameMode == ICS_Movelist ) )
    {
        manPtr = Logic->current[fromPtr].ManPtr;
        if( manPtr == Null )
            return false;
        /* Make the move */
        Logic->chessMove = chessMove;
        if( Logic->move() == false )
            return false;
        chessMove = Logic->chessMove;
    }

    /* Play sound for Promotion */
    if( QString( chessMove.SAN ).contains( QChar( '=' ) ) )
    {
        soundType = Sound::Promote;
    }

    /* Check to see if the game is ended */
    /* 50 Move Rule? */
    if( Logic->MoveCounter == 50 )
    {
        result = CMD_Result_Draw;
    }
    /*
                A Draw via 3 move repeat

    if( Record->isThreeMoveDraw() )
    {
        result = CMD_Result_Draw;
        Record->TAG_Result = "1/2-1/2";
        emit setStatusBar( GAME_DRAW );
    }      */
    /*
                A Draw?

                Condition 1 = White Material Draw
                Condition 2 = Black Material Draw
                Condition 3 = Both sides agree to a draw
    */
    if( ( Logic->OnMove == BLACK && Logic->isDraw( WHITE ) ) ||
            ( Logic->OnMove == WHITE && Logic->isDraw( BLACK ) ) ||
            ( Draw_Offered[WHITE] && Draw_Offered[BLACK] ) )
    {
        result = CMD_Result_Draw;
    }

    /* Is White under check? */
    if( Logic->isCheck( WHITE ) )
    {
        if( !Logic->isLegal( WHITE ) )
        {
            SAN += '#';
            result = CMD_Result_Black;
        }
        else
        {
            SAN += '+';
            soundType = Sound::Check;
        }
    }

    /* Is Black under check? */
    if( Logic->isCheck( BLACK ) )
    {
        if( !Logic->isLegal( BLACK ) )
        {
            SAN += '#';
            result = CMD_Result_White;
        }
        else
        {
            SAN += '+';
            soundType = Sound::Check;
        }
    }

    /* Check to make sure this isn't the starting position in an ICS match */
    if( QString( chessMove.SAN ) != "none" )
    {
        if( !SAN.isEmpty() )
        {
            strcat( chessMove.SAN, SAN.latin1() );
        }
        Record->history->add( chessMove, Logic->board(), Clock->getCentiseconds( WHITE ), Clock->getCentiseconds( BLACK ) );
        Record->currentIndex = Record->history->count( MatchHistory::Move ) - 1;
        if( JustMoved == true )
        {
            /* Send this move to engines and servers */
            emit sendCMD( Command( myID, CMD_Move, centiseconds( WHITE ), centiseconds( BLACK ),
                                   chessMove, Record->notation(false)->join(QString(" ")) ) );
        }
        /* Draw The Move */
        if( ICSGameMode != ICS_Movelist )
            Board->drawMove( chessMove );
    }

    /* Take care of changing turns, status messages, etc. */
    if( result == Null )
    {
        Clock->Moved();
        if( Paused )
            Clock->Pause();

        Logic->OnMove = !Logic->OnMove;
        /* Set Status Bar */
        if( Logic->OnMove == WHITE )
        {
            emit setStatusBar( WHITE_TURN );
        }
        if( Logic->OnMove == BLACK )
        {
            emit setStatusBar( BLACK_TURN );
        }

        /* Set Cursor */
        if( ( inputOnMove(true) == PLAYERLOCAL ) && ( inputOnMove() != PLAYERLOCAL ) )
        {
            Board->setCursor( myResource->CURSOR_Thinking );
        }
        else
        {
            Board->setCursor( myResource->CURSOR_Standard );
        }

        /* Deprecieate Draw Offers */
        if( Draw_Offered[WHITE] )
        {
            --Draw_Offered[WHITE];
        }
        if( Draw_Offered[BLACK] )
        {
            --Draw_Offered[BLACK];
        }

        /* Display NAG Anotation */
        if( chessMove.NAG != 0 )
        {
            emit setStatusBar( COMMENT + chessMove.NAG );
        }
    }
    else
    {
        /* End of Match */
        Board->setCursor( myResource->CURSOR_Standard );
        recvCMD( Command( myID, result ) );
        emit sendCMD( Command( myID, result ) );
        soundType = Sound::MatchOver;
    }
    Board->setNotation( Record->notation() );

    /* Unless we're getting a batch of moves... */
    if( ICSGameMode != ICS_Movelist )
    {
        Board->redrawLights();
        playSound( soundType );
    }

    /* Make the 'Retract' option available */
    if( ( inputOnMove() == PLAYERLOCAL ) && ( inputOnMove(true) == PLAYERPC ) && Record->history->count( MatchHistory::Move ) )
    {
        m_match->setItemEnabled( MENU_RETRACT, true );
    }
    else
    {
        m_match->setItemEnabled( MENU_RETRACT, false );
    }

    /* Handle special cases */
    if( loading() )
    {
        loadTimer = 1; // Loading...
    }
    else
    {
      setModified( true );
      Record->refresh();
    }
    if( preMoved )
    {
        preMoved = false;
        Board->setPremovePositions( Null, Null );
        chessMove = preMove;
        JustMoved = true;
        move();
    }

    /* Mail new move */
    if( ( Record->tags.find("Mode") == "EM" ) && ( !loading() ) )
    {
        Clock->Pause(); // Email
        int mailMove = KMessageBox::questionYesNo( 0, i18n("Would you like to send this move?"), i18n("Send Email?") );

        if( mailMove == KMessageBox::Yes )
        {
            Record->sendEmail( Logic->OnMove );
            Logic->OnMove = -1;
        }
        else
        {
            /* Restore the GUI */
            retract();
            Logic->OnMove = !Logic->OnMove;
            if( Logic->OnMove == WHITE )
            {
                setStatusMessage( WHITE_TURN );
            }
            else
            {
                setStatusMessage( BLACK_TURN );
            }
            if( ( inputOnMove(true) == PLAYERLOCAL ) && ( inputOnMove() != PLAYERLOCAL ) )
            {
                Board->setCursor( myResource->CURSOR_Thinking );
            }
            else
            {
                Board->setCursor( myResource->CURSOR_Standard );
            }
            Board->setNotation( Record->notation() );
        }
    }

    return true;
}
///////////////////////////////////////
//
//      match::retract
//
///////////////////////////////////////
void match::retract( void )
{
    emit setStatusBar(ILLEGAL_MOVE);
    review( Record->history->count( MatchHistory::Move ) - 2 );
    _retract();
}
void match::_retract( void )
{
    unsigned int oldCount = Record->history->count( MatchHistory::Move );
    if( oldCount )
    {
        while( oldCount == Record->history->count( MatchHistory::Move ) )
        {
            Record->history->undo();
        }
        Record->currentIndex = Record->history->count( MatchHistory::Move ) - 1;
    }
}
///////////////////////////////////////
//
//      match::review
//
///////////////////////////////////////
void match::review( int Index )
{
    int tmp(0), step;
    QString tmpS;
    if( ICSGameMode == ICS_Examine )
    {
        /* Examined Game: Use Examine Semantics */
        if( Index < (signed)Record->currentIndex )
            step = -1;
        else
            step = 1;
        if( ( Record->currentIndex == 0 ) && ( Index == 0 ) && ( step == 1 ) )
        {
            emit sendCMD( Command( myID, CMD_Examine_Forward ) );
        }
        if( step == 1 )
        {
            for( tmp = Record->currentIndex; tmp != Index; tmp += step )
            {
                emit sendCMD( Command( myID, CMD_Examine_Forward ) );
            }
        }
        else
        {
            for( tmp = Record->currentIndex; tmp != Index; tmp += step )
            {
                _retract();
                _retract();
                emit sendCMD( Command( myID, CMD_Examine_Backward ) );
            }
        }
    }
    else
    {
        /* Regular Game: Use Standard Review Semantics */
        if( Record->history->count( MatchHistory::Move ) == 0 )
            return;
        if( Index < 0 )
            return;
        if( Index > ( (signed int)Record->history->count( MatchHistory::Move ) - 1 ) )
            return;

        MatchHistory::MatchEvent event = Record->history->fetch( Index, MatchHistory::Move );

        /* Review differently depending on our Animation Option */
        if( myResource->OPTION_Animate_Moves )
        {
            if( (unsigned)Index > Record->currentIndex )
            {
                /* Forward */
                event = Record->history->fetch( Index - 1, MatchHistory::Move );
                Logic->setBoard( event.wholeBoard );
                while( tmp < _ARRAY_SIZE_ )
                {
                    Board->drawPosition( tmp++ );
                }
                event = Record->history->fetch( Index, MatchHistory::Move );
                Logic->setBoard( event.wholeBoard );
                Board->drawMove( event.move );
            }
            else if( (unsigned)Index < Record->currentIndex )
            {
                /* Reverse */
                event = Record->history->fetch( Index + 1, MatchHistory::Move );
                Logic->setBoard( event.wholeBoard );
                while( tmp < _ARRAY_SIZE_ )
                {
                    Board->drawPosition( tmp++ );
                }
                Board->drawMove( event.move, true );
                event = Record->history->fetch( Index, MatchHistory::Move );
                Logic->setBoard( event.wholeBoard );
            }
        }
        else
        {
            /* No Animation */
            Logic->setBoard( event.wholeBoard );
            while( tmp < _ARRAY_SIZE_ )
            {
                Board->drawPosition( tmp++ );
            }
            Board->drawMove( event.move );
        }

        /* Statusbar shows NAG if available, otherwise Army to move */
        if( event.move.NAG )
        {
            tmp = StatusCode;
            tmpS = StatusMessage;
            emit setStatusBar( COMMENT + event.move.NAG );
            StatusCode = tmp;
            StatusMessage = tmpS;
        }
        else
            emit setStatusBar( StatusCode, StatusMessage );
        Record->currentIndex = Index;
    }
    Board->setNotation( Record->notation(), Index );
}
///////////////////////////////////////
//
//  match::playSound
//
///////////////////////////////////////
void match::playSound( const int snd )
{
    if( ( !myResource->OPTION_Audio_Current_Only || Current ) && !Loading )
    {
        myResource->play( snd );
    }
}
///////////////////////////////////////
//
//  match::drawPosition
//
///////////////////////////////////////
void match::drawPosition( const int position )
{
  if( !Loading )
  {
    Board->drawPosition( position );
  }
}
///////////////////////////////////////
//
//  match::setModified
//
///////////////////////////////////////
void match::setModified( const bool &State )
{
  myFeatureList.EnableSave = State;
  emit updateWidgetFeatureList( myFeatureList );
}
