/***************************************************************************
                          pgn.cpp  -  description
                             -------------------
    begin                : Mon Jul 30 2001
    copyright            : (C) 2003 by Troy Corbin Jr.
    email                : tcorbin@users.sourceforge.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <sys/utsname.h>
#include <sys/types.h>
#include <qdatetime.h>
#include <qregexp.h>
#include <kprinter.h>
#include <krun.h>
#include <kfiledialog.h>
#include <kmessagebox.h>
#include <qfontmetrics.h>
#include <qpaintdevicemetrics.h>
#include "pgn.h"
#include "tabmanager.h"
#include "tab_pgnview.h"
#include "matchhistory.h"
#include "definitions.h"

pgn::pgn( resource *Rsrc, match_param *param )
{
    Resource = Rsrc;
    Param = param;
    pgnView = NULL;
    history = new MatchHistory();

    if( Param != NULL )
    {
        connect( Param, SIGNAL( valuesChanged() ), this, SLOT( parseMatchParam() ) );
    }
}
pgn::~pgn()
{
    if( Param )
    {
        delete Param;
    }
    if( pgnView )
    {
        if( Resource->tabManager->isTab( pgnView ) )
        {
            Resource->tabManager->removeTab( pgnView );
        }
    }
    delete history;
}
///////////////////////////////////////
//
//      pgn::parseMatchParam
//
///////////////////////////////////////
void pgn::parseMatchParam( void )
{
    if( Param == NULL )
    {
        return;
    }

    tags.replace( "White", Param->name( WHITE ) );
    tags.replace( "WhiteNA", Param->email( WHITE ) );
    tags.replace( "Black", Param->name( BLACK ) );
    tags.replace( "BlackNA", Param->email( BLACK ) );

    switch( Param->type( BLACK ) )
    {
    case PLAYERLOCAL:
        tags.replace( "BlackType", "human" );
        break;
    case PLAYERPC:
        tags.replace( "BlackType", "program" );
        break;
    case PLAYERTCP:
        tags.replace( "BlackType", "unknown" );
        tags.replace( "Mode", "ICS" );
        break;
    case PLAYEREMAIL:
        tags.replace( "BlackType", "unknown" );
        tags.replace( "Mode", "EM" );
        break;
    default:
        tags.replace( "BlackType", "unknown" );
        break;
    }
    switch( Param->type( WHITE ) )
    {
    case PLAYERLOCAL:
        tags.replace( "WhiteType", "human" );
        break;
    case PLAYERPC:
        tags.replace( "WhiteType", "program" );
        break;
    case PLAYERTCP:
        tags.replace( "WhiteType", "unknown" );
        tags.replace( "Mode", "ICS" );
        break;
    case PLAYEREMAIL:
        tags.replace( "WhiteType", "unknown" );
        tags.replace( "Mode", "EM" );
        break;
    default:
        tags.replace( "WhiteType", "unknown" );
        break;
    }

    switch( Param->getVariation() )
    {
    case GameType::Suicide:
        tags.replace( "RulesVariation", "Suicide" );
        break;
    case GameType::CrazyHouse:
        tags.replace( "RulesVariation", "Crazyhouse" );
        break;
    case GameType::BugHouse:
        tags.replace( "RulesVariation", "Bughouse" );
        break;
    case GameType::Standard: // Fall Through
    default:
        tags.replace( "RulesVariation", "Standard" );
        break;
    }

    if( Param->rating( WHITE ) != "" )
    {
        tags.replace( "WhiteELO", Param->rating( WHITE ) );
    }
    if( Param->rating( BLACK ) != "" )
    {
        tags.replace( "BlackELO", Param->rating( BLACK ) );
    }
}
///////////////////////////////////////
//
//      pgn::notation
//
///////////////////////////////////////
QStringList* pgn::notation( const int format )
{
    QStringList *list;
    MatchHistory::MatchEvent event;
    QString notation;
    QString SAN;
    unsigned int tmp(0), loop(0), Line(0);
    bool showLineNumber( false );

    list = new QStringList;

    switch( format )
    {
    case 0:
        /* CAN Raw move data ( ie. for UCI engines ) */
        for( loop = 0; loop < history->count(); loop++ )
        {
            event = history->fetch(loop);
            if( event.type == MatchHistory::Move )
                list->append( QString( event.move.CAN ) );
        }
        break;

    case 1:
        /* SAN For display to the user */
        for( loop = 0; loop < history->count(); loop++ )
        {
            event = history->fetch(loop);
            if( event.type == MatchHistory::Move )
            {
                Line = ( tmp + 2 ) >> 1;
                if( ( tmp % 2 ) == 0 )
                {
                    notation = QString( "%1. %2" ).arg( Line ).arg( event.move.SAN );
                }
                else
                {
                    notation = QString( "%1... %2" ).arg( Line ).arg( event.move.SAN );
                }
                list->append( notation );
                tmp++;
            }
        }
        break;

    case 2:
    case 3:
        /* SAN For PGN */
        /* 3 Adds hyperlinks */
        for( loop = 0; loop < history->count(); loop++ )
        {
            event = history->fetch(loop);
            if( event.type == MatchHistory::Move )
            {
                if( format == 3 )
                {
                  SAN = QString( "<a href=\"pgn://%1\">%2</a>" ).arg( loop ).arg( event.move.SAN );
                }
                else
                {
                  SAN = event.move.SAN;
                }
                Line = ( tmp + 2 ) >> 1;
                if( ( tmp % 2 ) == 0 )
                {
                    notation = QString( "%1. %2" ).arg( Line ).arg( SAN );
                }
                else
                {
                    if( showLineNumber )
                    {
                        notation = QString( "%1... %2" ).arg( Line ).arg( SAN );
                    }
                    else
                    {
                        notation = QString( SAN );
                    }
                }
                showLineNumber = false;
                /* Insert NAGs */
                if( event.move.NAG != 0 )
                {
                    notation += QString( " $%1" ).arg( QString::number( event.move.NAG ) );
                }
                /* Insert RAVs */
                if( !event.RAV.isEmpty() )
                {
                    notation += QString( " (%1)" ).arg( event.RAV );
                    showLineNumber = true;
                }
                /* Insert Annotations */
                if( !event.comment.isEmpty() )
                {
                    notation += QString( " {%1}" ).arg( event.comment );
                    showLineNumber = true;
                }
                list->append( notation );
                tmp++;
            }
        }
        break;

    default:
        break;
    }
    return list;
}
///////////////////////////////////////
//
//      pgn::caption
//
///////////////////////////////////////
QString pgn::caption( void )
{
    QString white = tags.find("White");
    QString black = tags.find("Black");

    if( Param->rating( WHITE ) != "" )
    {
        white += QString("(%1)").arg( tags.find("WhiteELO") );
    }
    if( Param->rating( BLACK ) != "" )
    {
        black += QString("(%1)").arg( tags.find("BlackELO") );
    }

    QString caption = i18n( "%1 vs. %2").arg( white ).arg( black );
    return caption;
}
///////////////////////////////////////
//
//      pgn::clear
//
///////////////////////////////////////
void pgn::clear( void )
{
    history->clear();
    whiteTCP.clear();
    blackTCP.clear();
    whiteTime = 300;
    blackTime = 300;
    currentIndex = 0;
    CurrentURL = "";
    Move_Data.clear();
    tags.clear();
}
///////////////////////////////////////
//
//      pgn::init
//
///////////////////////////////////////
void pgn::init( void )
{
    QString temp;
    struct utsname unamePtr;
    QDateTime qdt;

    clear();
    parseMatchParam();
    uname( &unamePtr );

    /* Build Date */
    qdt = QDateTime::currentDateTime();
    QString date = QString("%1.%2.%3").arg(qdt.date().year(),4).arg(qdt.date().month(),2).arg(qdt.date().day(),2);
    date.replace( QRegExp("\\s"), QString("0") );

    tags.replace( "Date", date );
    tags.replace( "Time", qdt.time().toString() );
    tags.replace( "Site", unamePtr.nodename );
    tags.replace( "Event", "Knights Computer Chess Game" );
    tags.replace( "Round", "-" );
    tags.replace( "Result", "*" );
}
///////////////////////////////////////
//
//      pgn::scan
//
///////////////////////////////////////
int pgn::scan( void )
{
    register int Section(0);
    QChar c;

    tags.clear();
    File_Position = File.at();

    /* Toplevel parsing loop */
    while( 1 )
    {
        /* Is this the end of the .pgn file? */
        if( Input.atEnd() )
        {
            close();
            return 100; // 100% Complete
        }
        currentLine = Input.readLine();
        if( currentLine.isEmpty() )
        {
            if( Section == 1 )
                return (int)( ( (float)File.at() / (float)File.size() ) * 100.0 );
        }
        c = getch();
        while( c != QChar::null )
        {
            switch( c )
            {
                /* Tag Pair */
            case '[':
                parseTag();
                break;
            case '%': /* Fall through */
            case ';':
                c = QChar::null;
                continue;
                break;
            default:
                Section = 1;
                c = QChar::null;
                continue;
                break;
            }
            c = getch();
        }
    }
    /* We should NEVER reach this point */
    close();
    return -1;
}
///////////////////////////////////////
//
//      pgn::load
//
///////////////////////////////////////
bool pgn::load( const int pos )
{
    QString Token, Value;
    QChar c;

    File.at(pos);
    clear();

    /* Toplevel parsing loop */
    while( 1 )
    {
        if( Input.atEnd() )
            break;
        currentLine = Input.readLine();

        if( currentLine.isEmpty() )
        {
            /*
                Finished getting TAGs... now adjust the match_param,
                start the notation tab, and continue to load the moves.
            */
            File_Position = File.at();

            /* Get and set variation */
            if( tags.find("RulesVariation") == "Bughouse" )
            {
                Param->setVariation( GameType::BugHouse );
            }
            else if( tags.find("RulesVariation") == "Crazyhouse" )
            {
                Param->setVariation( GameType::CrazyHouse );
            }
            else if( tags.find("RulesVariation") == "Suicide" )
            {
                Param->setVariation( GameType::Suicide );
            }
            else
            {
                // Default to Standard
                Param->setVariation( GameType::Standard );
            }

            Param->setName( WHITE, tags.find("White") );
            Param->setName( BLACK, tags.find("Black") );

            /* Finally, check for Email Chess Game */
            if( tags.find("Mode") == "EM" )
            {
                Param->setType(WHITE, PLAYERLOCAL);
                Param->setType(BLACK, PLAYERLOCAL);
            }

            /* Allocate the Tab_PGNView */
            pgnView = new tab_pgnView( this, Resource );
            connect( pgnView, SIGNAL( destroyed() ), this, SLOT( childViewDestroyed() ) );
            connect( pgnView, SIGNAL( displayInternalURL( const QString& ) ), this, SLOT( parseURL( const QString& ) ) );
            Resource->tabManager->addTab( pgnView, caption() );
            currentLine = "";

            return true;
        } // End if( currentLine.isEmpty() )

        c = getch();
        while( c != QChar::null )
        {
            Token = "";
            Value = "";
            switch( c )
            {
                /* Special break... look for the Knights tag */
            case '%':
                Value = getword();
                if( Value == "KNIGHTS_CMD" )
                    parseKnightsData();
                /* Fall through */
            case ';':
                c = QChar::null;
                continue;
                break;
                /* Tag Pair */
            case '[':
                parseTag();
                break;
            default:
                c = QChar::null;
                break;
            }
            c = getch();
        }
    }
    close();
    return true;
}
///////////////////////////////////////
//
//      pgn::loadNext
//
///////////////////////////////////////
bool pgn::loadNext( void )
{
    QChar c;
    QString Value;
    ChessMove Move;
    MatchHistory::MatchEvent event;
    bool postMove( false );

    /* Toplevel parsing loop */
    while( 1 )
    {
        if( currentLine.isEmpty() )
        {
            if( Input.atEnd() )
                break;
            currentLine = Input.readLine();
        }
        c = getch();
        while( c != QChar::null )
        {
            Value = "";
            switch( c )
            {
                /* Special break... look for the Knights tag */
            case '%':
                Value = getword();
                if( Value == "KNIGHTS_CMD" )
                    parseKnightsData();
                /* Fall through */
            case ';':
                c = QChar::null;
                continue;
                /* Tag Pair...next game, so we're done. */
            case '[':
                return false;
                /* Numeric Annotation Glyph */
            case '$':
                c = getch();
                while( ( c >= '0' ) && ( c <= '9' ) )
                {
                    Value += c;
                    c = getch();
                }
                event = history->last();
                event.move.NAG = Value.toInt();
                history->undo();
                history->add( event );
                break;
                /* Recursive Annotation Variations */
            case '(':
                parseRAV();
                break;
                /* Textual Annotation */
            case '{':
                parseAnnotation();
                break;
                /* Reserved for future expansion by the PGN */
            case '<':
                while( c != '>' )
                    c = getch();
                break;
                /* Everything from '0' to '*' is null to us and falls through */
            case '0':
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
            case ' ':
            case '/':
            case '-':
            case '.':
            case '*':
                break;
            default:
                if( postMove )
                {
                    currentLine.prepend( c );
                    return true;
                }
                /* Catch Standard Algebraic Notation */
                while(  ( c != QChar::null ) &&
                        ( c != ' '         ) &&
                        ( c != ';'         ) &&
                        ( c != '%'         ) )
                {
                    Value += c;
                    c = getch();
                }
                strcpy( Move.SAN, Value.latin1() );
                emit processMove( Move );
                postMove = true;
            }
            c = getch();
        }
    }
    return false;
}
///////////////////////////////////////
//
//      pgn::save
//
///////////////////////////////////////
bool pgn::save( QString& URL )
{
    QFile *Save = new QFile( URL );

    close();
    CurrentURL = URL;
    if( Resource->OPTION_Reuse_PGN && ( URL == Resource->PGN_Filename ) )
    {
        if( !Save->open( IO_WriteOnly | IO_Append ) )
        {
            KMessageBox::sorry( 0, i18n("Knights could not open the file to be saved."), i18n("Problem saving file.") );
            delete Save;
            return false;
        }
    }
    else
    {
        if( !Save->open( IO_WriteOnly ) )
        {
            KMessageBox::sorry( 0, i18n("Knights could not open the file to be saved."), i18n("Problem saving file.") );
            delete Save;
            return false;
        }
    }

    QString Token, Value;
    QTextStream *Output = new QTextStream( Save );

    if( ( tags.find("Type") == "EM" ) && ( tags.find("CmailGameName").isEmpty() ) )
    {
        tags.replace( "CmailGameName", CurrentURL.section( '/', -1 ) );
    }
    tags.replace( "PlyCount", QString().setNum( history->count( MatchHistory::Move ) ) );
    for( unsigned int loop = 0; loop < tags.count(); loop++ )
    {
        (*Output) << "[" << tags.keyAt(loop) << " \"" << tags[loop] << "\"]\n";
    }

    /* Save internal data if this game is unfinished */
    if( tags.find("Result") == "*" )
    {
        (*Output) << "% KNIGHTS_CMD WhiteType ";
        switch( Param->type(WHITE) )
        {
        case PLAYERPC:
            (*Output) << "PC\n";
            break;
        case PLAYERTCP:
            (*Output) << "TCP\n";
            break;
        case PLAYEREMAIL:
            (*Output) << "Email\n";
            break;
        case PLAYERLOCAL:
        default:
            (*Output) << "Local\n";
            break;
        }
        (*Output) << "% KNIGHTS_CMD BlackType ";
        switch( Param->type(BLACK) )
        {
        case PLAYERPC:
            (*Output) << "PC\n";
            break;
        case PLAYERTCP:
            (*Output) << "TCP\n";
            break;
        case PLAYEREMAIL:
            (*Output) << "Email\n";
            break;
        case PLAYERLOCAL:
        default:
            (*Output) << "Local\n";
            break;
        }
        (*Output) << "% KNIGHTS_CMD WhiteTime " << whiteTime << "\n";
        (*Output) << "% KNIGHTS_CMD BlackTime " << blackTime << "\n";
        (*Output) << "% KNIGHTS_CMD DONE\n";
    }
    /* End of internal data save */

    (*Output) << "\n";
    QStringList *list = notation(2);
    QString SAN = list->join( " " );
    delete list;

    //  kdWarning() << SAN.right( 20 ) << endl;

    unsigned int pos = 80;
    unsigned int lastPos = 0;
    while( pos < SAN.length() )
    {
        while( SAN.at( pos ) != ' ' )
            pos--;
        SAN = SAN.replace( pos, 1, QString("\n") );
        lastPos = pos;
        pos = lastPos + 80;
    }
    //  kdWarning() << SAN.right( 20 ) << endl;

    (*Output) << SAN << " " << tags.find("Result") << "\n\n";

    Save->close();

    delete Output;
    delete Save;
    return true;
}
///////////////////////////////////////
//
//      pgn::getch
//
///////////////////////////////////////
QChar pgn::getch( void )
{
    QChar c;

    c = currentLine.at(0);
    currentLine.remove( 0, 1 );
    return c;
}
///////////////////////////////////////
//
//      pgn::getword
//
///////////////////////////////////////
QString pgn::getword( void )
{
    QChar c;
    QString word;

    do
    {
        c = getch();
    }
    while( c == ' ' );

    while( ( c != ' ' ) && ( c != QChar::null ) )
    {
        word += c;
        c = getch();
    }
    return word;
}
///////////////////////////////////////
//
//      pgn::open
//
///////////////////////////////////////
bool pgn::open( const QString &URL )
{
    close();
    if( !KIO::NetAccess::download( URL, tempFile, NULL ) )
        return false;

    File.setName( tempFile );
    if( !File.open( IO_ReadOnly ) )
    {
        close();
        return false;
    }
    Input.setDevice( &File );
    CurrentURL = URL;
    File.at(0);
    return true;
}
///////////////////////////////////////
//
//      pgn::close
//
///////////////////////////////////////
void pgn::close( void )
{
    if( !File.isOpen() )
    {
        File.close();
    }
    if( !tempFile.isEmpty() )
    {
        KIO::NetAccess::removeTempFile( tempFile );
        tempFile = "";
    }
}
///////////////////////////////////////
//
//      pgn::parseTag
//
///////////////////////////////////////
void pgn::parseTag( void )
{
    QChar c;
    QString Token;
    QString Value;

    c = getch();
    while( c == ' ' )
        c = getch();
    while( c != '\"' )
    {
        if( c == ' ' )
        {
            c = getch();
            continue;
        }
        Token += c;
        c = getch();
    }
    c = getch();
    while( c != '\"' )
    {
        Value += c;
        c = getch();
    }
    c = getch();
    while( c != ']' )
        c = getch();
    /* Now Apply the Token/Value that we got */
    tags.replace( Token, Value );
}
///////////////////////////////////////
//
//      pgn::Parse_Annotation
//
///////////////////////////////////////
void pgn::parseAnnotation( void )
{
    QChar c;
    QString string;
    MatchHistory::MatchEvent event;

    c = getch();
    while( c != '}' )
    {
        if( c == QChar::null )
        {
            if( Input.eof() )
                break;
            currentLine = Input.readLine();
            c = ' ';
        }
        string += c;
        c = getch();
    }
    event = history->last();
    event.comment = string;
    history->undo();
    history->add( event );
}
///////////////////////////////////////
//
//      pgn::Parse_RAV
//
///////////////////////////////////////
void pgn::parseRAV( void )
{
    int RAVLevel(1);
    QString string;
    QChar c;
    MatchHistory::MatchEvent event;

    while( RAVLevel )
    {
        c = getch();
        if( c == QChar::null )
        {
            if( Input.eof() )
                break;
            currentLine = Input.readLine();
            c = ' ';
        }
        if( c == ')' )
        {
            RAVLevel--;
            if( !RAVLevel )
                break;
        }
        if( c == '(' )
        {
            RAVLevel++;
        }
        string += c;
    }
    event = history->last();
    event.RAV = string;
    history->undo();
    history->add( event );
}
///////////////////////////////////////
//
//      pgn::parseKnightsData
//
///////////////////////////////////////
void pgn::parseKnightsData( void )
{
    QString Key;
    QString Value;

    Key = getword();
    if( Key == "DONE" )
    {
        emit processSpecial();
        return;
    }
    if( Key == "WhiteType" )
    {
        Value = getword();
        if( Value == "Local" )
            Param->setType(WHITE, PLAYERLOCAL);
        if( Value == "PC" )
            Param->setType(WHITE, PLAYERPC);
        if( Value == "TCP" )
            Param->setType(WHITE, PLAYERTCP);
        if( Value == "Email" )
            Param->setType(WHITE, PLAYEREMAIL);
        return;
    }
    if( Key == "BlackType" )
    {
        Value = getword();
        if( Value == "Local" )
            Param->setType(BLACK, PLAYERLOCAL);
        if( Value == "PC" )
            Param->setType(BLACK, PLAYERPC);
        if( Value == "TCP" )
            Param->setType(BLACK, PLAYERTCP);
        if( Value == "Email" )
            Param->setType(BLACK, PLAYEREMAIL);
        return;
    }
    if( Key == "WhiteTime" )
    {
        Value = getword();
        whiteTime = Value.toInt();
        return;
    }
    if( Key == "BlackTime" )
    {
        Value = getword();
        blackTime = Value.toInt();
        return;
    }
}
///////////////////////////////////////
//
//      pgn::print
//
///////////////////////////////////////
void pgn::print( void )
{
    if( !pgnView )
    {
        /* Allocate the Tab_PGNView */
        pgnView = new tab_pgnView( this, Resource );
        Resource->tabManager->addTab( pgnView, caption() );
        connect( pgnView, SIGNAL( destroyed() ), this, SLOT( childViewDestroyed() ) );
        connect( pgnView, SIGNAL( displayInternalURL( const QString& ) ), this, SLOT( parseURL( const QString& ) ) );
        pgnView->refresh();
    }
    pgnView->print();
}
///////////////////////////////////////
//
//      pgn::refresh
//
///////////////////////////////////////
void pgn::refresh( void )
{
  if( pgnView )
  {
    pgnView->refresh();
  }
}
///////////////////////////////////////
//
//  pgn::sendEmail
//
///////////////////////////////////////
bool pgn::sendEmail( const int &onMove )
{
    QString cmdLine = Resource->Email_Command_Line;
    if( cmdLine.isEmpty() )
    {
        KMessageBox::sorry( 0, i18n("No email application has been configured to work with Knights."), i18n("No Email Client Setup") );
        return false;
    }

    /* Save PGN */
    QString URL = KFileDialog::getSaveFileName( QString::null, Resource->FileFilter, 0, i18n( "Save Match..." ) );
    if( URL.isNull() )
    {
        return false;
    }
    if( URL.findRev( ".pgn", -1, false ) == -1 )
    {
        URL += ".pgn";
    }
    save( URL );

    /* Replace macros with data */
    cmdLine.replace( "%s", QString("\"Chess Move: %1 vs %2\"").arg( tags.find("White") ).arg( tags.find("Black") ) );
    cmdLine.replace( "%m", QString("\"%1\"").arg(URL) );
    if( onMove == WHITE )
    {
        cmdLine.replace( "%a", QString("\"%1\"").arg( tags.find("WhiteNA") ) );
    }
    else
    {
        cmdLine.replace( "%a", QString("\"%1\"").arg( tags.find("BlackNA") ) );
    }
    kdWarning() << "pgn::sendEmail: Running: " << cmdLine << endl;
    KRun::runCommand( cmdLine );
    return true;
}
///////////////////////////////////////
//
//      pgn::childViewDestroyed
//
///////////////////////////////////////
void pgn::childViewDestroyed( void )
{
    pgnView = NULL;
}
///////////////////////////////////////
//
//      pgn::parseURL
//
///////////////////////////////////////
void pgn::parseURL( const QString &URL )
{
    if( !URL.startsWith( "pgn://" ) )
    {
      return;
    }
    QString myURL = URL.right( URL.length() - 6 );
    myURL = myURL.remove( '/' );
    emit requestReview( myURL.toInt() );
}
///////////////////////////////////////
//
//      pgn::getNAG
//
///////////////////////////////////////
QString pgn::getNAG( int num )
{
    QString Line;

    switch( num )
    {
    case 1:
        Line = i18n( "Good move" );
        break;
    case 2:
        Line = i18n( "Poor move" );
        break;
    case 3:
        Line = i18n( "Very good move" );
        break;
    case 4:
        Line = i18n( "Very poor move" );
        break;
    case 5:
        Line = i18n( "Speculative move" );
        break;
    case 6:
        Line = i18n( "Questionable move" );
        break;
    case 7:
        Line = i18n( "Forced move" );
        break;
    case 8:
        Line = i18n( "Singular move" );
        break;
    case 9:
        Line = i18n( "Worst move" );
        break;
    case 10:
        Line = i18n( "Drawish position" );
        break;
    case 11:
        Line = i18n( "Equal chances, quiet position" );
        break;
    case 12:
        Line = i18n( "Equal chances, active position" );
        break;
    case 13:
        Line = i18n( "Unclear position" );
        break;
    case 14:
        Line = i18n( "White has a slight advantage" );
        break;
    case 15:
        Line = i18n( "Black has a slight advantage" );
        break;
    case 16:
        Line = i18n( "White has a moderate advantage" );
        break;
    case 17:
        Line = i18n( "Black has a moderate advantage" );
        break;
    case 18:
        Line = i18n( "White has a decisive advantage" );
        break;
    case 19:
        Line = i18n( "Black has a decisive advantage" );
        break;
    case 20:
        Line = i18n( "White has a crushing advantage ( Black should resign )" );
        break;
    case 21:
        Line = i18n( "Black has a crushing advantage ( White should resign )" );
        break;
    case 22:
        Line = i18n( "White is in zugzwang" );
        break;
    case 23:
        Line = i18n( "Black is in zugzwang" );
        break;
    case 24:
        Line = i18n( "White has a slight space advantage" );
        break;
    case 25:
        Line = i18n( "Black has a slight space advantage" );
        break;
    case 26:
        Line = i18n( "White has a moderate space advantage" );
        break;
    case 27:
        Line = i18n( "Black has a moderate space advantage" );
        break;
    case 28:
        Line = i18n( "White has a decisive space advantage" );
        break;
    case 29:
        Line = i18n( "Black has a decisive space advantage" );
        break;
    case 30:
        Line = i18n( "White has a slight time ( development ) advantage" );
        break;
    case 31:
        Line = i18n( "Black has a slight time ( development ) advantage" );
        break;
    case 32:
        Line = i18n( "White has a moderate time ( development ) advantage" );
        break;
    case 33:
        Line = i18n( "Black has a moderate time ( development ) advantage" );
        break;
    case 34:
        Line = i18n( "White has a decisive time ( development ) advantage" );
        break;
    case 35:
        Line = i18n( "Black has a decisive time ( development ) advantage" );
        break;
    case 36:
        Line = i18n( "White has the initiative" );
        break;
    case 37:
        Line = i18n( "Black has the initiative" );
        break;
    case 38:
        Line = i18n( "White has a lasting initiative" );
        break;
    case 39:
        Line = i18n( "Black has a lasting initiative" );
        break;
    case 40:
        Line = i18n( "White has the attack" );
        break;
    case 41:
        Line = i18n( "Black has the attack" );
        break;
    case 42:
        Line = i18n( "White has insufficient compensation for material deficit" );
        break;
    case 43:
        Line = i18n( "Black has insufficient compensation for material deficit" );
        break;
    case 44:
        Line = i18n( "White has sufficient compensation for material deficit" );
        break;
    case 45:
        Line = i18n( "Black has sufficient compensation for material deficit" );
        break;
    case 46:
        Line = i18n( "White has more than adequate compensation for material deficit" );
        break;
    case 47:
        Line = i18n( "Black has more than adequate compensation for material deficit" );
        break;
    case 48:
        Line = i18n( "White has a slight center control advantage" );
        break;
    case 49:
        Line = i18n( "Black has a slight center control advantage" );
        break;
    case 50:
        Line = i18n( "White has a moderate center control advantage" );
        break;
    case 51:
        Line = i18n( "Black has a moderate center control advantage" );
        break;
    case 52:
        Line = i18n( "White has a decisive center control advantage" );
        break;
    case 53:
        Line = i18n( "Black has a decisive center control advantage" );
        break;
    case 54:
        Line = i18n( "White has a slight kingside control advantage" );
        break;
    case 55:
        Line = i18n( "Black has a slight kingside control advantage" );
        break;
    case 56:
        Line = i18n( "White has a moderate kingside control advantage" );
        break;
    case 57:
        Line = i18n( "Black has a moderate kingside control advantage" );
        break;
    case 58:
        Line = i18n( "White has a decisive kingside control advantage" );
        break;
    case 59:
        Line = i18n( "Black has a decisive kingside control advantage" );
        break;
    case 60:
        Line = i18n( "White has a slight queenside control advantage" );
        break;
    case 61:
        Line = i18n( "Black has a slight queenside control advantage" );
        break;
    case 62:
        Line = i18n( "White has a moderate queenside control advantage" );
        break;
    case 63:
        Line = i18n( "Black has a moderate queenside control advantage" );
        break;
    case 64:
        Line = i18n( "White has a decisive queenside control advantage" );
        break;
    case 65:
        Line = i18n( "Black has a decisive queenside control advantage" );
        break;
    case 66:
        Line = i18n( "White has a vulnerable first rank" );
        break;
    case 67:
        Line = i18n( "Black has a vulnerable first rank" );
        break;
    case 68:
        Line = i18n( "White has a well protected first rank" );
        break;
    case 69:
        Line = i18n( "Black has a well protected first rank" );
        break;
    case 70:
        Line = i18n( "White has a poorly protected king" );
        break;
    case 71:
        Line = i18n( "Black has a poorly protected king" );
        break;
    case 72:
        Line = i18n( "White has a well protected king" );
        break;
    case 73:
        Line = i18n( "Black has a well protected king" );
        break;
    case 74:
        Line = i18n( "White has a poorly placed king" );
        break;
    case 75:
        Line = i18n( "Black has a poorly placed king" );
        break;
    case 76:
        Line = i18n( "White has a well placed king" );
        break;
    case 77:
        Line = i18n( "Black has a well placed king" );
        break;
    case 78:
        Line = i18n( "White has a very weak pawn structure" );
        break;
    case 79:
        Line = i18n( "Black has a very weak pawn structure" );
        break;
    case 80:
        Line = i18n( "White has a moderately weak pawn structure" );
        break;
    case 81:
        Line = i18n( "Black has a moderately weak pawn structure" );
        break;
    case 82:
        Line = i18n( "White has a moderately strong pawn structure" );
        break;
    case 83:
        Line = i18n( "Black has a moderately strong pawn structure" );
        break;
    case 84:
        Line = i18n( "White has a very strong pawn structure" );
        break;
    case 85:
        Line = i18n( "Black has a very strong pawn structure" );
        break;
    case 86:
        Line = i18n( "White has poor knight placement" );
        break;
    case 87:
        Line = i18n( "Black has poor knight placement" );
        break;
    case 88:
        Line = i18n( "White has good knight placement" );
        break;
    case 89:
        Line = i18n( "Black has good knight placement" );
        break;
    case 90:
        Line = i18n( "White has poor bishop placement" );
        break;
    case 91:
        Line = i18n( "Black has poor bishop placement" );
        break;
    case 92:
        Line = i18n( "White has good bishop placement" );
        break;
    case 93:
        Line = i18n( "Black has good bishop placement" );
        break;
    case 94:
        Line = i18n( "White has poor rook placement" );
        break;
    case 95:
        Line = i18n( "Black has poor rook placement" );
        break;
    case 96:
        Line = i18n( "White has good rook placement" );
        break;
    case 97:
        Line = i18n( "Black has good rook placement" );
        break;
    case 98:
        Line = i18n( "White has poor queen placement" );
        break;
    case 99:
        Line = i18n( "Black has poor queen placement" );
        break;
    case 100:
        Line = i18n( "White has good queen placement" );
        break;
    case 101:
        Line = i18n( "Black has good queen placement" );
        break;
    case 102:
        Line = i18n( "White has poor piece coordination" );
        break;
    case 103:
        Line = i18n( "Black has poor piece coordination" );
        break;
    case 104:
        Line = i18n( "White has good piece coordination" );
        break;
    case 105:
        Line = i18n( "Black has good piece coordination" );
        break;
    case 106:
        Line = i18n( "White has played the opening very poorly" );
        break;
    case 107:
        Line = i18n( "Black has played the opening very poorly" );
        break;
    case 108:
        Line = i18n( "White has played the opening poorly" );
        break;
    case 109:
        Line = i18n( "Black has played the opening poorly" );
        break;
    case 110:
        Line = i18n( "White has played the opening well" );
        break;
    case 111:
        Line = i18n( "Black has played the opening well" );
        break;
    case 112:
        Line = i18n( "White has played the opening very well" );
        break;
    case 113:
        Line = i18n( "Black has played the opening very well" );
        break;
    case 114:
        Line = i18n( "White has played the middlegame very poorly" );
        break;
    case 115:
        Line = i18n( "Black has played the middlegame very poorly" );
        break;
    case 116:
        Line = i18n( "White has played the middlegame poorly" );
        break;
    case 117:
        Line = i18n( "Black has played the middlegame poorly" );
        break;
    case 118:
        Line = i18n( "White has played the middlegame well" );
        break;
    case 119:
        Line = i18n( "Black has played the middlegame well" );
        break;
    case 120:
        Line = i18n( "White has played the middlegame very well" );
        break;
    case 121:
        Line = i18n( "Black has played the middlegame very well" );
        break;
    case 122:
        Line = i18n( "White has played the ending very poorly" );
        break;
    case 123:
        Line = i18n( "Black has played the ending very poorly" );
        break;
    case 124:
        Line = i18n( "White has played the ending poorly" );
        break;
    case 125:
        Line = i18n( "Black has played the ending poorly" );
        break;
    case 126:
        Line = i18n( "White has played the ending well" );
        break;
    case 127:
        Line = i18n( "Black has played the ending well" );
        break;
    case 128:
        Line = i18n( "White has played the ending very well" );
        break;
    case 129:
        Line = i18n( "Black has played the ending very well" );
        break;
    case 130:
        Line = i18n( "White has slight counterplay" );
        break;
    case 131:
        Line = i18n( "Black has slight counterplay" );
        break;
    case 132:
        Line = i18n( "White has moderate counterplay" );
        break;
    case 133:
        Line = i18n( "Black has moderate counterplay" );
        break;
    case 134:
        Line = i18n( "White has decisive counterplay" );
        break;
    case 135:
        Line = i18n( "Black has decisive counterplay" );
        break;
    case 136:
        Line = i18n( "White has moderate time control pressure" );
        break;
    case 137:
        Line = i18n( "Black has moderate time control pressure" );
        break;
    case 138:
        Line = i18n( "White has severe time control pressure" );
        break;
    case 139:
        Line = i18n( "Black has severe time control pressure" );
    case 140:
        Line = i18n( "With the idea..." );
        break;
    case 141:
        Line = i18n( "Aimed against..." );
        break;
    case 142:
        Line = i18n( "Better Move" );
        break;
    case 143:
        Line = i18n( "Worse Move" );
        break;
    case 144:
        Line = i18n( "Equivalent move" );
        break;
    case 145:
        Line = i18n( "Editor's Remark" );
        break;
    case 146:
        Line = i18n( "Novelty" );
        break;
    case 147:
        Line = i18n( "Weak point" );
        break;
    case 148:
        Line = i18n( "Endgame" );
        break;
    case 149:
        Line = i18n( "Line" );
        break;
    case 150:
        Line = i18n( "Diagonal" );
        break;
    case 151:
        Line = i18n( "White has a pair of Bishops" );
        break;
    case 152:
        Line = i18n( "Black has a pair of Bishops" );
        break;
    case 153:
        Line = i18n( "Bishops of opposite color" );
        break;
    case 154:
        Line = i18n( "Bishops of same color" );
        break;
    case 190:
        Line = i18n( "Etc." );
        break;
    case 191:
        Line = i18n( "Doubled pawns" );
        break;
    case 192:
        Line = i18n( "Isolated pawn" );
        break;
    case 193:
        Line = i18n( "Connected pawns" );
        break;
    case 194:
        Line = i18n( "Hanging pawns" );
        break;
    case 195:
        Line = i18n( "Backwards pawn" );
        break;
    default:
        Line = "";
        break;
    }
    return Line;
}

