/***************************************************************************
                          tabbox.cpp  -  description
                             -------------------
    begin                : Fri Sep 13 2002
    copyright            : (C) 2003 by Troy Corbin Jr.
    email                : tcorbin@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tabbox.moc"
#include "tabpage.h"
#include "tabmanager.h"
#include "resource.h"
#include "accel.h"
#include "menubar.h"
#include "command.h"
#include "list_pgn.h"
#include "dlg_newmatch.h"
#include "knightspixcache.h"
#include "dlg_settings.h"
#include "core.h"
#include <klocale.h>
#include <kkeydialog.h>
#include <kfiledialog.h>
#include <kpopupmenu.h>
#include <qstyle.h>
#include <qtabwidget.h>

///////////////////////////////////////
//
//  TabBox::constructor
//
///////////////////////////////////////
TabBox::TabBox( resource *rsrc ) : QVBox(0,"TabBox",Qt::WDestructiveClose)
{
    myResource = rsrc;
    minimized = false;
    setMargin( QApplication::style().defaultFrameWidth() );
    setFocusPolicy( ClickFocus );

    myMenuBar = new MenuBar( this, "myMenuBar", myResource );
    connect( myMenuBar, SIGNAL( selected(int) ), this, SLOT( menuAction(int) ) );

    myTabs = new QTabWidget( this, "myTabs" );
    myTabs->setTabShape( QTabWidget::Rounded );
    myTabs->setTabPosition( QTabWidget::Top );
    connect( myTabs, SIGNAL( currentChanged( QWidget* ) ), this, SLOT( currentChanged( QWidget* ) ) );

    myAccel = new Accel( this, myResource->myAccel );
    connect( this, SIGNAL( focus( const QChar& ) ), myResource->myAccel, SIGNAL( focus( const QChar& ) ) );
}
///////////////////////////////////////
//
//  TabBox::destructor
//
///////////////////////////////////////
TabBox::~TabBox()
{
    shutdown();
    delete myAccel;
    delete myMenuBar;

    QCustomEvent *e = new QCustomEvent( EVENT_Close_Tab );
    e->setData( this );
    QApplication::postEvent( myResource->tabManager, e );
}
///////////////////////////////////////
//
//      TabBox::shutdown
//
///////////////////////////////////////
bool TabBox::shutdown( void )
{
    while( myTabs->count() )
    {
        if( !removeTab( myTabs->page(0), true ) )
        {
            return false;
        }
    }
    return true;
}
///////////////////////////////////////
//
//      TabBox::keyPressEvent
//
///////////////////////////////////////
void TabBox::keyPressEvent( QKeyEvent *e )
{
    QChar input;

    if( ( e->state() | Qt::ShiftButton ) == Qt::ShiftButton )
    {
        input = e->text().at(0);
        if( input.isLetterOrNumber() )
        {
            emit focus( input );
            e->accept();
            return;
        }
    }
    e->ignore();
}
///////////////////////////////////////
//
//      TabBox::addTab
//
///////////////////////////////////////
void TabBox::addTab( QWidget *child, const QString &caption )
{
    if( QString( child->className() ) == "TabPage" )
    {
        /* We were passed the TabPage, not the actual widget */
        myTabs->addTab( child, caption );
        myTabs->showPage( child );
        changeMyCaption( child );
        connect( child, SIGNAL( newParent( TabBox* ) ), this, SIGNAL( newTabBox( TabBox* ) ) );
        connect( child, SIGNAL( requestDestruction() ), this, SLOT( destroyChild() ) );
        currentChanged( child );
    }
    else
    {
        TabPage *newPage = new TabPage( (QWidget*)this, child, myResource );
        newPage->setCaption( caption );

        myTabs->addTab( newPage, caption );
        myTabs->showPage( newPage );
        changeMyCaption( newPage );
        connect( newPage, SIGNAL( newParent( TabBox* ) ), this, SIGNAL( newTabBox( TabBox* ) ) );
        connect( newPage, SIGNAL( requestDestruction() ), this, SLOT( destroyChild() ) );
        currentChanged( newPage );
    }
}
///////////////////////////////////////
//
//      TabBox::removeTab
//
///////////////////////////////////////
bool TabBox::removeTab( QWidget *child, bool deleteChild )
{
    QWidget *tabPage;
    QWidget *trueWidget;

    if( QString( child->className() ) == "TabPage" )
    {
        /* We were passed the TabPage, not the actual widget */
        tabPage = child;
        trueWidget = ((TabPage*)child)->getChild();
    }
    else
    {
        tabPage = child->parentWidget();
        trueWidget = child;
    }

    /* Delete it if requested */
    if( deleteChild )
    {
        /* Attempt to close the child widget */
        if( ((TabPage*)tabPage)->closeChild() )
        {
          emit saveTabGeometry( QString( trueWidget->className() ), size() );
          disconnect( tabPage, SIGNAL( newParent( TabBox* ) ), this, SIGNAL( newTabBox( TabBox* ) ) );
          disconnect( tabPage, SIGNAL( requestDestruction() ), this, SLOT( destroyChild() ) );
            myTabs->removePage( tabPage );
            delete child;
            updateWindowMenu();
            return true;
        }
    }
    else
    {
      emit saveTabGeometry( QString( trueWidget->className() ), size() );
      disconnect( tabPage, SIGNAL( newParent( TabBox* ) ), this, SIGNAL( newTabBox( TabBox* ) ) );
      disconnect( tabPage, SIGNAL( requestDestruction() ), this, SLOT( destroyChild() ) );
        myTabs->removePage( tabPage );
        updateWindowMenu();
        return true;
    }
    return false;
}
///////////////////////////////////////
//
//  TabBox::showTab
//
///////////////////////////////////////
void TabBox::showTab( QWidget *child )
{
    if( isChild( child ) )
    {
        myTabs->showPage( child->parentWidget() );
    }
}
///////////////////////////////////////
//
//      TabBox::destroyChild
//
///////////////////////////////////////
void TabBox::destroyChild( void )
{
    QObject *child = ((QObject*)sender());
    QCustomEvent *e = NULL;

    /* Be careful... make sure we have a valid child calling us */
    if( child == NULL )
        return;
    if( !child->isWidgetType() )
        return;
    if( !isChild( ((QWidget*)child) ) )
        return;

    e = new QCustomEvent( EVENT_Close_Tab );
    e->setData( child );
    QApplication::postEvent( myResource->tabManager, e );
}
///////////////////////////////////////
//
//  TabBox::changeMyCaption
//
///////////////////////////////////////
void TabBox::changeMyCaption( QWidget *child )
{
    setCaption( i18n("%1 - Knights").arg( myTabs->tabLabel( child ) ) );
}
///////////////////////////////////////
//
//  TabBox::count
//
///////////////////////////////////////
int TabBox::count( void )
{
    return myTabs->count();
}
///////////////////////////////////////
//
//      TabBox::isChild
//
///////////////////////////////////////
bool TabBox::isChild( QWidget *child )
{
    for( int tmp=0; tmp < myTabs->count(); tmp++ )
    {
        if( myTabs->page(tmp) == child )
            return true;
        if( ((TabPage*)myTabs->page(tmp))->getChild() == child )
            return true;
    }
    return false;
}
///////////////////////////////////////
//
//      TabBox::getChildName
//
///////////////////////////////////////
QString TabBox::getChildName( const int &index )
{
    if( index > myTabs->count() )
    {
        return QString::null;
    }
    return QString( ((TabPage*)myTabs->page(index))->getChild()->className() );
}
///////////////////////////////////////
//
//      TabBox::changeCaption
//
///////////////////////////////////////
void TabBox::changeCaption( QWidget *child, const QString &caption )
{
    if( QString( child->className() ) == "TabPage" )
    {
        ((TabPage*)child)->setCaption( caption );
        myTabs->setTabLabel( child, caption );
    }
    else if( isChild( child ) )
    {
        ((TabPage*)child->parentWidget())->setCaption( caption );
        myTabs->setTabLabel( child->parentWidget(), caption );
    }
    changeMyCaption( myTabs->currentPage() );
}
///////////////////////////////////////
//
//      TabBox::currentChanged
//
///////////////////////////////////////
void TabBox::currentChanged( QWidget *page )
{
  changeMyCaption( page );

  /* Get the new Tab's Context Menu */
  KPopupMenu *contextMenu = ((TabPage*)page)->getMenu();
  QString menuCaption = ((TabPage*)page)->getMenuCaption();
  myMenuBar->changeContextMenu( contextMenu, menuCaption );

  /* Make sure we're tied into this widget's feature list */
  disconnect( this, SLOT( recieveWidgetFeatureList( const WFL& ) ) );
  connect( (TabPage*)page, SIGNAL( recievedWidgetFeatureList( const WFL& ) ), this, SLOT( recieveWidgetFeatureList( const WFL& ) ) );
  recieveWidgetFeatureList( ((TabPage*)page)->getWidgetFeatureList() );

  updateWindowMenu();
}
///////////////////////////////////////
//
//      TabBox::menuAction
//
///////////////////////////////////////
void TabBox::menuAction( int menuItem )
{
    list_pgn *listPGN = NULL;
    dlg_newmatch *newMatch = NULL;
    QCustomEvent *e = NULL;
    QString URL;

    switch( menuItem )
    {
      case MENU_CLOSE:
          e = new QCustomEvent( EVENT_Close_Tab );
          e->setData( myTabs->currentPage() );
          QApplication::postEvent( myResource->tabManager, e );
          break;

      case MENU_QUIT:
          /* Signal TabManager to instruct all TabBoxes to close all children */
          QApplication::postEvent( myResource->tabManager, new QCustomEvent( EVENT_Shutdown_App ) );
          break;

      case MENU_FULLSCREEN:
          if( isFullScreen() )
          {
            //myMenuBar->show();
            showNormal();
          }
          else
          {
            //myMenuBar->hide();
            showFullScreen();
          }
          break;

      case MENU_NEW_GAME:
          newMatch = new dlg_newmatch( this, "NewMatch", myResource );
          connect( newMatch, SIGNAL( initMatch( match_param* ) ), myResource->Core, SLOT( newMatch( match_param* ) ) );
          newMatch->show();
          break;

      case MENU_NEW_CONSOLE: // Drop Through
      case MENU_NEW_SEEK_LIST: // Drop Through
      case MENU_NEW_SEEK_GRAPH:
          myResource->Core->createNewTab( menuItem );
          break;

      case MENU_SETTINGS_DIALOG:
          SettingsDialog = new dlg_settings( this, "settings", myResource );
          connect( SettingsDialog, SIGNAL( redrawBoard() ), myResource->pixCache, SLOT( resetCache() ) );
          connect( SettingsDialog, SIGNAL( resetServer() ), myResource->Core, SLOT( resetServer() ) );
          SettingsDialog->show();
          break;

      case MENU_BINDINGS_DIALOG:
          KKeyDialog::configure( myAccel, false, this );
          break;

      case MENU_INSTALL_THEMES:
          myResource->installThemes();
          break;

      case MENU_CONNECT:
          if(myResource->Core->isOnline())
          {
              setCursor( myResource->CURSOR_Standard );
              QApplication::postEvent( myResource->Core, new QCustomEvent( EVENT_Del_IO_Net ) );
          }
          else
          {
              /*
                  By passing Null as the ID, we're telling core that we're
                  creating a new internetio, not reusing an exsisting one.
              */
              myResource->Core->createNewIO( PLAYERTCP, Null);
              myResource->tabManager->refreshMenus();
          }
          break;

      case MENU_LOAD:
          URL = KFileDialog::getOpenFileName( QString::null, myResource->FileFilter, 0, i18n( "Load PGN..." ) );
          if( !URL.isEmpty() )
          {
              listPGN = new list_pgn( 0, "listPGN" );
              myResource->tabManager->addTab( listPGN, URL.section( '/', -1 ) );
              connect( listPGN, SIGNAL( selected( const QString&, const int& ) ), myResource->Core, SLOT( load( const QString&, const int& ) ) );
              listPGN->setURL( URL );
          }
          break;

      default:
          /* Did User Select a New Tab to Display? */
          if( menuItem >= MENU_SELECT_WINDOW )
          {
              myTabs->setCurrentPage( menuItem - MENU_SELECT_WINDOW );
              break;
          }

          /* All unknown menu items are sent to the child, if possible */
          ((TabPage*)myTabs->currentPage())->recvCMD( Command( 0, menuItem ) );
          break;
    }
}
///////////////////////////////////////
//
//  TabBox::refreshMenu
//
///////////////////////////////////////
void TabBox::refreshMenu( void )
{
    myMenuBar->createMenus();
    currentChanged( (TabPage*)myTabs->currentPage() );
}
///////////////////////////////////////
//
//  TabBox::updateWindowMenu
//
///////////////////////////////////////
void TabBox::updateWindowMenu( void )
{
    QStringList windowList;
    int current = myTabs->currentPageIndex();

    for( int loop = 0; loop < myTabs->count(); loop++ )
    {
        windowList += myTabs->tabLabel( myTabs->page( loop ) );
    }

    myMenuBar->createWindowMenu( windowList, current );
}
///////////////////////////////////////
//
//  TabBox::recieveWidgetFeatureList
//
///////////////////////////////////////
void TabBox::recieveWidgetFeatureList( const WFL &list )
{
  myMenuBar->setItemEnabled( MENU_SAVE, list.EnableSave );
  myMenuBar->setItemEnabled( MENU_SAVEAS, list.EnableSaveAs );
  myMenuBar->setItemEnabled( MENU_PRINT, list.EnablePrint );
}
///////////////////////////////////////
//
//      TabBox::hideEvent
//
///////////////////////////////////////
void TabBox::hideEvent( QHideEvent* )
{
  if( myResource->OPTION_Pause_On_Minimize )
  {
    for( int tmp=0; tmp < myTabs->count(); tmp++ )
    {
        ((TabPage*)myTabs->page(tmp))->recvCMD( Command( 0, MENU_PAUSEALL ) );
    }
    minimized = true;
  }
}
///////////////////////////////////////
//
//      TabBox::showEvent
//
///////////////////////////////////////
void TabBox::showEvent( QShowEvent* )
{
  if( minimized )
  {
    for( int tmp=0; tmp < myTabs->count(); tmp++ )
    {
        ((TabPage*)myTabs->page(tmp))->recvCMD( Command( 0, MENU_PAUSEALL ) );
    }
    minimized = false;
  }
}
