/***************************************************************************
                          tabmanager.cpp  -  description
                             -------------------
    begin                : Sat Sep 14 2002
    copyright            : (C) 2003 by Troy Corbin Jr.
    email                : tcorbin@users.sf.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "tabmanager.moc"
#include "resource.h"
#include <kdebug.h>

static QString myConfigGroup( "TabManager" );

TabManager::TabManager( resource *rsrc ) : QObject()
{
    myResource = rsrc;
    appConfig = kapp->config();
    myList.setAutoDelete( false );
}

TabManager::~TabManager()
{
  TabBox *boxPtr = myList.first();
  while( boxPtr != 0 )
  {
    boxPtr->shutdown();
    boxPtr = myList.next();
  }
}

///////////////////////////////////////
//
//      TabManager::addTab
//
///////////////////////////////////////
void TabManager::addTab( QWidget *child, const QString &caption )
{
    TabBox *boxPtr;
    QSize defaultSize;
    appConfig->setGroup( myConfigGroup );

    /* Get default size */
    if( child->sizeHint().isValid() )
    {
        defaultSize = child->sizeHint();
				defaultSize.setHeight( defaultSize.height() + 64 ); // Factor in Tabs, which the widget won't do.
    }
    else
    {
        defaultSize = QSize( 380, 420 );
    }

    if( myList.count() == 0 )
    {
        /* First Tab always generates a new TabBox */
        boxPtr = new TabBox( myResource );
        manageNewBox( boxPtr );
        boxPtr->addTab( child, caption );

        /* Get Geometry from Saved Config */
        QSize newSize = appConfig->readSizeEntry( QString( child->className() ), &defaultSize );
        boxPtr->resize( newSize );
        boxPtr->show();
    }
    else
    {
        /* Compare relatives to see who we're most compatable with */
        int bestBox = 0;
        int bestScore = -100000;
        QStringList myRelations = appConfig->readListEntry( "_rel_" + QString( child->className() ) );

        /* Create a collective score for the whole box */
        for( unsigned int boxLoop = 0; boxLoop < myList.count(); boxLoop++ )
        {
            int currentScore = 0;
            boxPtr = myList.at( boxLoop );
            for( int pageLoop = 0; pageLoop < boxPtr->count(); pageLoop++ )
            {
                QString widgetName = boxPtr->getChildName( pageLoop );
                QStringList::Iterator IT = myRelations.find( widgetName );
                if( IT == myRelations.end() )
                {
                    /* No previous relation, score of zero */
                    continue;
                }
                else
                {
                    ++IT;
                    currentScore += (*IT).toInt();
                }
            }
            if( currentScore > bestScore )
            {
                bestScore = currentScore;
                bestBox = boxLoop;
            }
        }
        /* We use the best box, or make a new box if the best score is < 0 */
        if( bestScore < 0 )
        {
            /* A new TabBox */
            boxPtr = new TabBox( myResource );
            manageNewBox( boxPtr );
            boxPtr->addTab( child, caption );

            /* Get Geometry from Saved Config */
            QSize newSize = appConfig->readSizeEntry( QString( child->className() ), &defaultSize );
            boxPtr->resize( newSize );
            boxPtr->show();
        }
        else
        {
            boxPtr = myList.at( bestBox );
            boxPtr->addTab( child, caption );
        }
    }
}
///////////////////////////////////////
//
//      TabManager::changeCaption
//
///////////////////////////////////////
void TabManager::changeCaption( QWidget *child, const QString &caption )
{
    TabBox *boxPtr = getParentBox( child );

    if( boxPtr != NULL )
    {
        boxPtr->changeCaption( child, caption );
    }
}
///////////////////////////////////////
//
//      TabManager::removeTab
//
///////////////////////////////////////
void TabManager::removeTab( QObject *child )
{
  /*
      Remove TabBox Children Directly - Should only need this when the TabBox
      is already dead, so do not delete it, and don't access the pointer.
  */
  if( myList.find( ((TabBox*)child) ) != -1 )
  {
    myList.remove( ((TabBox*)child) );
    if( !myList.count() )
    {
      QApplication::postEvent( this, new QCustomEvent( EVENT_Shutdown_App ) );
    }
    return;
  }

  /* Uhh, this *IS* a child object, right? */
  if( isTab( (QWidget*)child ) )
  {

    /* Make sure this is a QWidget we can manipulate */
    if( !child->isWidgetType() )
    {
        kdError() << "TabManager::removeTab: Can not remove non-widget class " << child->className() << endl;
        return;
    }

    QString childName( ((QWidget*)child)->className() );
    TabBox *boxPtr = getParentBox( ((QWidget*)child) );

    /* Remove this Child's Parent, the TabPage */
    if( boxPtr != NULL )
    {
        boxPtr->removeTab( ((QWidget*)child), true );
        // Remove the Box too?
        if( boxPtr->count() == 0 )
        {
            myList.remove( boxPtr );
            delete boxPtr;

            /* Was this our last window? */
            if( !myList.count() )
            {
                QApplication::postEvent( this, new QCustomEvent( EVENT_Shutdown_App ) );
            }
        }
        return;
    }
    kdWarning() << "Could not find child: " << childName << endl;
  }
}
///////////////////////////////////////
//
//      TabManager::showTab
//
///////////////////////////////////////
void TabManager::showTab( QWidget *child )
{
    TabBox *boxPtr = getParentBox( child );
    if( boxPtr != NULL )
    {
        boxPtr->showTab( child );
    }
}
///////////////////////////////////////
//
//      TabManager::isTab
//
///////////////////////////////////////
bool TabManager::isTab( QWidget *child )
{
    if( getParentBox( child ) == NULL )
    {
        return false;
    }
    return true;
}
///////////////////////////////////////
//
//      TabManager::saveTabGeometry
//
///////////////////////////////////////
void TabManager::saveTabGeometry( const QString &childName, const QSize &childSize )
{
    /* Save This Widget's Size For Later */
    appConfig->setGroup( myConfigGroup );
    appConfig->writeEntry( childName, childSize );
    appConfig->sync();

    kdDebug() << "Saved geometry for Widget: " << childName << endl;
}
///////////////////////////////////////
//
//      TabManager::childEvent
//
///////////////////////////////////////
void TabManager::childEvent( QChildEvent *event )
{
    if( event->inserted() )
    {
        /* New Child */
        if( QString( event->child()->className() ) == "TabBox" )
        {
            manageNewBox( ((TabBox*)event->child()) );
        }
        else
        {
          addTab( ((QWidget*)event->child()), QString( "Untitled" ) );
        }
    }
}
///////////////////////////////////////
//
//      TabManager::event
//
///////////////////////////////////////
bool TabManager::event( QEvent *e )
{
    if( e->type() == EVENT_Shutdown_App )
    {
        /* User wants to shutdown */
        TabBox *boxPtr = myList.first();
        while( boxPtr != 0 )
        {
          if( boxPtr->shutdown() == false )
          {
            return false;
          }
          boxPtr = myList.next();
        }
        qApp->quit();
        return true;
    }
    else if( e->type() == EVENT_Close_Tab )
    {
        removeTab( ((QObject*)((QCustomEvent*)e)->data()) );
        return true;
    }
    return false;
}
///////////////////////////////////////
//
//      TabManager::manageNewBox
//
///////////////////////////////////////
void TabManager::manageNewBox( TabBox *newBox )
{
    myList.append( newBox );
    connect( newBox, SIGNAL( destroyed( QObject* ) ), this, SLOT( removeTab( QObject* ) ) );
    connect( newBox, SIGNAL( saveTabGeometry( const QString&, const QSize& ) ), this, SLOT( saveTabGeometry( const QString&, const QSize& ) ) );
    connect( newBox, SIGNAL( newTabBox( TabBox* ) ), this, SLOT( manageNewBox( TabBox* ) ) );
}
///////////////////////////////////////
//
//      TabManager::getParentBox
//
///////////////////////////////////////
TabBox* TabManager::getParentBox( QWidget *child )
{
    TabBox *boxPtr;
    for( unsigned int index=0; index < myList.count(); index++ )
    {
        boxPtr = myList.at(index);
        if( boxPtr->isChild( child ) )
        {
            return boxPtr;
        }
    }
    return NULL;
}
///////////////////////////////////////
//
//      TabManager::updateRelations
//
///////////////////////////////////////
void TabManager::updateRelations( QWidget *child, TabBox *relatives, const int &score )
{
    QString childName = QString( child->className() );

    appConfig->setGroup( myConfigGroup );
    QStringList myRelations = appConfig->readListEntry( "_rel_" + childName );

    /* Update all relations */
    for( int loop = 0; loop < relatives->count(); loop++ )
    {
        QString widgetName = relatives->getChildName( loop );
        QStringList hisRelations = appConfig->readListEntry( "_rel_" + widgetName );
        QStringList::Iterator IT = myRelations.find( widgetName );
        if( IT == myRelations.end() )
        {
            /* Add widget to my relations list */
            myRelations.append( widgetName );
            myRelations.append( QString::number( score ) );
            /* Add myself to his relations too! */
            hisRelations.append( childName );
            hisRelations.append( QString::number( score ) );
            appConfig->writeEntry( "_rel_" + widgetName, hisRelations );
        }
        else
        {
            /* Update my relations with this widget */
            ++IT;
            int oldScore = (*IT).toInt();
            IT = myRelations.remove( IT );
            myRelations.insert( IT, QString( "%1" ).arg( oldScore + score ) );
            /* Now update it from his end */
            IT = hisRelations.find( childName );
            ++IT;
            oldScore = (*IT).toInt();
            IT = hisRelations.remove( IT );
            hisRelations.insert( IT, QString( "%1" ).arg( oldScore + score ) );
            appConfig->writeEntry( "_rel_" + widgetName, hisRelations );
        }
    }

    /* Save relations */
    appConfig->writeEntry( "_rel_" + childName, myRelations );
    appConfig->sync();
}
///////////////////////////////////////
//
//      TabManager::refreshMenus
//
///////////////////////////////////////
void TabManager::refreshMenus( void )
{
    for( unsigned int loop = 0; loop < myList.count(); loop++ )
    {
        TabBox *boxPtr = myList.at( loop );
        boxPtr->refreshMenu();
    }
}
